from dataclasses import dataclass, field
from functools import wraps

import pyarrow as pa
from biocore import DataHandler
from biocore.utils.import_util import is_datasets_available

from biofit.processing import BaseProcessor, ProcessorConfig, SelectedColumnTypes
from biofit.utils import logging

logger = logging.get_logger(__name__)


@dataclass
class ModelConfig(ProcessorConfig):
    """Base class for feature extraction processor configurations."""

    _fit_process_desc: str = field(default="Fitting the model", init=False, repr=False)
    predict_process_desc: str = field(
        default="Predicting target output", init=False, repr=False
    )
    predict_proba_process_desc: str = field(
        default="Predicting target output probabilities", init=False, repr=False
    )
    processor_type: str = field(default="models", init=False, repr=False)
    _missing_val: float = field(default=0, init=False, repr=False)
    _missing_val_pa_type: pa.DataType = field(
        default=pa.float64(), init=False, repr=False
    )
    class_names: list = field(default=None, init=False, repr=False)

    task: str = None


class Model(BaseProcessor):
    """Base class for models."""

    @wraps(BaseProcessor._process_transform)
    def predict(
        self,
        X,
        input_columns: SelectedColumnTypes = None,
        keep_unused_columns: bool = None,
        raise_if_missing: bool = None,
        cache_output: bool = None,
        cache_dir: str = None,
        cache_file_name: str = None,
        load_from_cache_file: bool = None,
        batched: bool = None,
        batch_size: int = None,
        batch_format: str = None,
        output_format: str = None,
        map_kwargs: dict = None,
        num_proc: int = None,
        fingerprint: str = None,
    ):
        """Predict the model."""
        self._method_prefix = "_predict"
        self.output_dtype = "int64"
        self._input_columns = self._set_input_columns_and_arity(input_columns)
        return self._process_transform(
            X,
            keep_unused_columns=keep_unused_columns,
            raise_if_missing=raise_if_missing,
            cache_output=cache_output,
            cache_dir=cache_dir,
            cache_file_name=cache_file_name,
            load_from_cache_file=load_from_cache_file,
            batched=batched,
            batch_size=batch_size,
            batch_format=batch_format,
            output_format=output_format,
            map_kwargs=map_kwargs,
            num_proc=num_proc,
            fingerprint=fingerprint,
        )

    def predict_proba(
        self,
        X,
        input_columns: SelectedColumnTypes = None,
        keep_unused_columns: bool = None,
        raise_if_missing: bool = None,
        cache_output: bool = None,
        cache_dir: str = None,
        cache_file_name: str = None,
        load_from_cache_file: bool = None,
        batched: bool = None,
        batch_size: int = None,
        batch_format: str = None,
        output_format: str = None,
        map_kwargs: dict = None,
        num_proc: int = None,
        fingerprint: str = None,
    ):
        self._method_prefix = "_predict_proba"
        self.output_dtype = "float64"
        self._input_columns = self._set_input_columns_and_arity(input_columns)
        return self._process_transform(
            X,
            keep_unused_columns=keep_unused_columns,
            raise_if_missing=raise_if_missing,
            cache_output=cache_output,
            cache_dir=cache_dir,
            cache_file_name=cache_file_name,
            load_from_cache_file=load_from_cache_file,
            batched=batched,
            batch_size=batch_size,
            batch_format=batch_format,
            output_format=output_format,
            map_kwargs=map_kwargs,
            num_proc=num_proc,
            fingerprint=fingerprint,
        )

    @staticmethod
    def suggest_params() -> dict:
        """Get hyperparameters for an optuna trial."""
        raise NotImplementedError

    def _process_fit_input(self, X, **kwargs):
        target_col = kwargs["fn_kwargs"]["extra_indices"][0]
        if target_col and "class" in self.config.task.lower():
            if isinstance(target_col, list):
                target_col = target_col[0]

            target_col = DataHandler.get_column_names(X, generate_cols=True)[target_col]

            if self.config.n_classes is None:
                self.config.n_classes = DataHandler.nunique(X, target_col)

            if is_datasets_available():
                from datasets import Dataset

                if isinstance(X, Dataset):
                    feat = X._info.features[target_col]
                    self.config.class_names = feat.names

            if self.config.n_classes > 2:
                self.config.task = "multiclass_classification"
            else:
                self.config.task = "binary_classification"
        return super()._process_fit_input(X, **kwargs)

    def _process_fit_output(self, input, out):
        if hasattr(self, "classes_"):
            if self.config.class_names is not None and isinstance(
                self.classes_[0], int
            ):
                if sorted(self.classes_.tolist()) == list(range(self.config.n_classes)):
                    self.config.class_names = [
                        self.config.class_names[i] for i in self.classes_
                    ]
                else:
                    self.config.class_names = [str(i) for i in self.classes_]
            else:
                self.config.class_names = [str(i) for i in self.classes_]
        return super()._process_fit_output(input, out)

    def _process_transform_input(self, X, **kwargs):
        if self._method_prefix == "_predict_proba":
            self.config._n_features_out = self.config.n_classes
            self.config._feature_names_out = self.config.class_names
            self.output_dtype = "float64"
        else:
            if self.config.extra_names_in_ is not None:
                self.config._feature_names_out = self.config.extra_names_in_[0]
            self.config._n_features_out = 1
        return super()._process_transform_input(X, **kwargs)

    def _process_transform_batch_input(self, X, *fn_args, **fn_kwargs):
        X, args, kwargs = super()._process_transform_batch_input(
            X, *fn_args, **fn_kwargs
        )
        if DataHandler.supports_named_columns(X):
            input_cols = DataHandler.get_column_names(X)
            missing_cols = set(self.config.feature_names_in_) - set(input_cols)
            intersecting_cols = set(input_cols) & set(self.config.feature_names_in_)

            X_dims = DataHandler.get_shape(X)
            num_cols = None
            if len(X_dims) > 1:
                num_cols = X_dims[1]

            num_non_existing = num_cols - len(intersecting_cols)
            if num_non_existing:
                logger.warning_once(
                    f"Dataset has {num_non_existing} out of {num_cols} columns that were "
                    "not in the training data. Dropping these columns."
                )
                X = DataHandler.select_columns(X, list(intersecting_cols))

            if missing_cols:
                if self.config._missing_val is None:
                    self.config._missing_val_str = "`None`"
                elif self.config._missing_val == 0:
                    self.config._missing_val_str = "zeroes"
                else:
                    self.config._missing_val_str = f"{self.config._missing_val}"
                logger.warning_once(
                    f"Dataset is missing {len(missing_cols)} out of "
                    f"{len(self.config.feature_names_in_)} columns that were in the "
                    "training data. Adding these columns as "
                    f"{self.config._missing_val_str}."
                )
                num_rows = DataHandler.get_shape(X)[0]
                zeros_mat = pa.table(
                    {
                        col: pa.array(
                            [self.config._missing_val] * num_rows,
                            type=self.config._missing_val_pa_type,
                        )
                        for col in missing_cols
                    }
                )
                X = DataHandler.concat(
                    [X, DataHandler.to_format(zeros_mat, DataHandler.get_format(X))],
                    axis=1,
                )
            # Reorder columns to match the training data
            X = DataHandler.select_columns(X, self.config.feature_names_in_)

        return X, args, kwargs



# ruff: noqa
from .min_prevalence_sample_filter import (
    MinPrevalenceSampleFilter,
    MinPrevalenceRowSampleFilterConfig,
    MinPrevalenceRowSampleFilterConfigForOTU,
    MinPrevalenceRowSampleFilterConfigForSNP,
    MinPrevalenceRowSampleFilterConfigForMaldi,
)
from .plot_min_prevalence_sample_filter import (
    MinPrevalenceRowPlotterConfig,
    MinPrevalenceRowPlotterConfigForMetagenomics,
    MinPrevalenceRowPlotterConfigForOTU,
    MinPrevalenceRowPlotterConfigForSNP,
    MinPrevalenceRowPlotterConfigForASV,
    MinPrevalenceRowPlotterConfigForMaldi,
    MinPrevalenceRowPlotterConfigForGenomics,
    MinPrevalenceRowPlotterConfigForReadCount,
    MinPrevalenceRowPlotterConfigForProteomics,
    MinPrevalenceRowPlotter,
)



# ruff: noqa
from .min_prevalence_feature_selector import (
    MinPrevalenceFeatureSelector,
    MinPrevalenceFeatureSelectorConfig,
    MinPrevalenceFeatureSelectorConfigForMetagenomics,
    MinPrevalenceFeatureSelectorConfigForOTU,
    MinPrevalenceFeatureSelectorConfigForSNP,
)
from .plot_min_prevalence_feature_selector import (
    MinPrevalenceFeatureSelectorPlotter,
    MinPrevalencePlotterConfig,
    MinPrevalencePlotterConfigForASV,
    MinPrevalencePlotterConfigForGenomics,
    MinPrevalencePlotterConfigForMetagenomics,
    MinPrevalencePlotterConfigForOTU,
    MinPrevalencePlotterConfigForProteomics,
    MinPrevalencePlotterConfigForMaldi,
    MinPrevalencePlotterConfigForReadCount,
    MinPrevalencePlotterConfigForSNP,
)



from dataclasses import dataclass, field
from typing import TYPE_CHECKING, List, Optional, Union

import numpy as np
from sklearn.linear_model import Lasso

from biofit.integration.biosets import get_feature
from biofit.processing import SelectedColumnTypes, sync_backup_config
from biofit.utils import logging

from ..logistic_regression import LogisticRegressionModel
from ..models import Model, ModelConfig

logger = logging.get_logger(__name__)

if TYPE_CHECKING:
    pass


@dataclass
class LassoConfig(ModelConfig):
    _fit_input_feature_types: List[Union[None, type]] = field(
        default_factory=lambda: [None, get_feature("TARGET_FEATURE_TYPES")],
        init=False,
        repr=False,
    )
    _fit_unused_feature_types: List[Union[None, type]] = field(
        default_factory=lambda: [get_feature("METADATA_FEATURE_TYPES"), None],
        init=False,
        repr=False,
    )
    _transform_unused_feature_types: List[type] = field(
        default_factory=lambda: [get_feature("METADATA_FEATURE_TYPES")],
        init=False,
        repr=False,
    )
    _fit_process_desc: str = field(
        default="Fitting the lasso model", init=False, repr=False
    )
    predict_process_desc: str = field(
        default=None,
        init=False,
        repr=False,
    )
    predict_proba_process_desc: str = field(
        default=None,
        init=False,
        repr=False,
    )
    processor_name: str = field(default="lasso", init=False, repr=False)

    input_columns: SelectedColumnTypes = None
    target_column: SelectedColumnTypes = None
    alpha = 1.0
    fit_intercept: bool = True
    precompute: Union[bool, np.ndarray] = False
    copy_X: bool = True
    max_iter: int = 1000
    tol: float = 1e-4
    warm_start: bool = False
    positive: bool = False
    random_state: Optional[int] = 42
    class_weight: Union[dict, str] = None
    selection: str = "cyclic"  # "cyclic" or "random"
    solver: str = "saga"

    use_predict_proba: bool = False
    task: str = None
    estimator: Lasso = field(default=None, init=False, repr=False)
    n_classes: int = None

    def __post_init__(self):
        if self.task is None:
            self.task = "classification" if self.use_predict_proba else "regression"
        self._fit_process_desc = f"Fitting the lasso {self.task} model"
        self.predict_process_desc = f"Predicting with the lasso {self.task} model"
        self.predict_proba_process_desc = (
            f"Predicting probabilities with the lasso {self.task} model"
        )


class LassoConfigForOTU(LassoConfig):
    dataset_name: str = field(default="otu", init=False, repr=False)
    log_transform: Union[str, bool] = field(default="log2_1p", init=False, repr=False)


class LassoModel(Model):
    config_class = LassoConfig
    config: LassoConfig
    lasso: Lasso

    @sync_backup_config
    def set_params(self, **kwargs):
        self.config = self.config.replace_defaults(**kwargs)
        self.lasso = Lasso(
            alpha=self.config.alpha,
            fit_intercept=self.config.fit_intercept,
            precompute=self.config.precompute,
            copy_X=self.config.copy_X,
            max_iter=self.config.max_iter,
            tol=self.config.tol,
            warm_start=self.config.warm_start,
            positive=self.config.positive,
            random_state=self.config.random_state,
            selection=self.config.selection,
        )
        return self

    @property
    def feature_importances_(self):
        return self.config.estimator.coef_.flatten()

    @property
    def feature_names_in_(self):
        return self.config.feature_names_in_

    def fit(
        self,
        X,
        y=None,
        input_columns: SelectedColumnTypes = None,
        target_column: SelectedColumnTypes = None,
        raise_if_missing: bool = None,
        cache_output: bool = None,
        load_from_cache_file: bool = None,
        batched: bool = True,
        batch_size: int = 1000,
        batch_format: str = None,
        num_proc: int = None,
        map_kwargs: dict = {"fn_kwargs": {}},
        cache_dir: str = None,
        cache_file_name: str = None,
        fingerprint: str = None,
    ) -> "LassoModel":
        self.config._input_columns = self._set_input_columns_and_arity(
            input_columns, target_column
        )
        return self._process_fit(
            X,
            y,
            raise_if_missing=raise_if_missing,
            cache_output=cache_output,
            load_from_cache_file=load_from_cache_file,
            batched=batched,
            batch_size=batch_size,
            batch_format=batch_format,
            num_proc=num_proc,
            map_kwargs=map_kwargs,
            cache_dir=cache_dir,
            cache_file_name=cache_file_name,
            fingerprint=fingerprint,
        )

    def predict(
        self,
        X,
        input_columns: SelectedColumnTypes = None,
        keep_unused_columns: bool = None,
        raise_if_missing: bool = None,
        cache_output: bool = None,
        cache_dir: str = None,
        cache_file_name: str = None,
        load_from_cache_file: bool = None,
        batched: bool = None,
        batch_size: int = None,
        batch_format: str = None,
        output_format: str = None,
        map_kwargs: dict = None,
        num_proc: int = None,
        fingerprint: str = None,
    ):
        self._input_columns = self._set_input_columns_and_arity(input_columns)
        return self._process_transform(
            X,
            keep_unused_columns=keep_unused_columns,
            raise_if_missing=raise_if_missing,
            cache_output=cache_output,
            cache_dir=cache_dir,
            cache_file_name=cache_file_name,
            load_from_cache_file=load_from_cache_file,
            batched=batched,
            batch_size=batch_size,
            batch_format=batch_format,
            output_format=output_format,
            map_kwargs=map_kwargs,
            num_proc=num_proc,
            fingerprint=fingerprint,
        )

    def predict_proba(
        self,
        X,
        input_columns: SelectedColumnTypes = None,
        keep_unused_columns: bool = None,
        raise_if_missing: bool = None,
        cache_output: bool = None,
        cache_dir: str = None,
        cache_file_name: str = None,
        load_from_cache_file: bool = None,
        batched: bool = None,
        batch_size: int = None,
        batch_format: str = None,
        output_format: str = None,
        map_kwargs: dict = None,
        num_proc: int = None,
        fingerprint: str = None,
    ):
        self._input_columns = self._set_input_columns_and_arity(input_columns)
        return self._process_transform(
            X,
            keep_unused_columns=keep_unused_columns,
            raise_if_missing=raise_if_missing,
            cache_output=cache_output,
            cache_dir=cache_dir,
            cache_file_name=cache_file_name,
            load_from_cache_file=load_from_cache_file,
            batched=batched,
            batch_size=batch_size,
            batch_format=batch_format,
            output_format=output_format,
            map_kwargs=map_kwargs,
            num_proc=num_proc,
            fingerprint=fingerprint,
        )

    def _fit_sklearn(self, X, y):
        if self.config.transform_log == "log2_1p":
            X = np.log2(X + 1)
        self.config.estimator = self.lasso.fit(X, y)
        return self

    def _predict_sklearn(self, X):
        return self.config.estimator.predict(X)

    def _predict_proba_sklearn(self, X):
        logger.warning_once(
            "predict_proba is not supported for Lasso. Returning using predict instead."
        )
        return self._predict_sklearn(X)

    @staticmethod
    def suggest_params(data):
        return {
            "alpha": (
                "suggest_float",
                (
                    1e-8,
                    1e3,
                ),
                {"log": True},
            ),
            "fit_intercept": ("suggest_categorical", ([True, False],), {}),
            "max_iter": (
                "suggest_int",
                (
                    1000,
                    10000,
                ),
                {},
            ),
        }

    @staticmethod
    def suggest_first_trial():
        return {
            "alpha": 1.0,
            "fit_intercept": True,
            "max_iter": 1000,
        }


# Lasso for Classification is a logistic regression model with L1 penalty, except
# that with the additional layer of converting the scores for classes to the "winning"
# class output label.
class LassoForClassification(LogisticRegressionModel):
    class_config = LassoConfig

    def __init__(
        self,
        alpha=1.0,
        fit_intercept: bool = True,
        copy_X: bool = True,
        max_iter: int = 1000,
        tol: float = 1e-4,
        warm_start: bool = False,
        class_weight: str = None,
        random_state: int = 42,
        solver: str = "saga",
        config: LassoConfig = None,
        **kwargs,
    ):
        if "C" in kwargs:
            alpha = 1.0 / kwargs.pop("C")
        super().__init__(
            C=1.0 / alpha,  # C is the inverse of alpha
            fit_intercept=fit_intercept,
            max_iter=max_iter,
            tol=tol,
            penalty="l1",
            solver=solver,
            warm_start=warm_start,
            class_weight=class_weight,
            random_state=random_state,
            config=config,
        )

    @sync_backup_config
    def set_params(self, **kwargs):
        if "alpha" in kwargs:
            kwargs["C"] = 1.0 / kwargs.pop("alpha")
        super().set_params(**kwargs)
        return self

    @property
    def classes_(self):
        return self.config.estimator.classes_

    @staticmethod
    def suggest_params(data):
        params = {
            "C": (
                "suggest_float",
                (
                    1e-4,
                    1e2,
                ),
                {"log": True},
            ),
            "tol": (
                "suggest_float",
                (
                    1e-5,
                    1e-2,
                ),
                {"log": True},
            ),
            "fit_intercept": ("suggest_categorical", ([True, False],), {}),
            "max_iter": ("suggest_categorical", ([100, 200, 500, 1000],), {}),
            "class_weight": ("suggest_categorical", (["balanced", "None"],), {}),
        }
        return params

    @staticmethod
    def suggest_first_trial():
        return {
            "alpha": 1.0,
            "tol": 1e-4,
            "fit_intercept": True,
            "max_iter": 1000,
            "class_weight": "None",
        }


class LassoForRegression(LassoModel):
    pass



# Copyright 2022 The HuggingFace Team. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Generic utilities
"""

import inspect
import tempfile
import contextlib
import shutil
import os
import stat
from pathlib import Path
from collections import OrderedDict, UserDict
from collections.abc import MutableMapping
from contextlib import ExitStack, contextmanager
from dataclasses import fields, is_dataclass
from enum import Enum
from typing import Any, ContextManager, List, Tuple, Union, Optional, Generator

import numpy as np

# optional imports
try:
    import polars as pl
except ImportError:
    pl = None


class cached_property(property):
    """
    Descriptor that mimics @property but caches output in member variable.

    From tensorflow_datasets

    Built-in in functools from Python 3.8.
    """

    def __get__(self, obj, objtype=None):
        # See docs.python.org/3/howto/descriptor.html#properties
        if obj is None:
            return self
        if self.fget is None:
            raise AttributeError("unreadable attribute")
        attr = "__cached_" + self.fget.__name__
        cached = getattr(obj, attr, None)
        if cached is None:
            cached = self.fget(obj)
            setattr(obj, attr, cached)
        return cached


def _set_write_permission_and_retry(func, path, excinfo):
    os.chmod(path, stat.S_IWRITE)
    func(path)


@contextlib.contextmanager
def SoftTemporaryDirectory(
    suffix: Optional[str] = None,
    prefix: Optional[str] = None,
    dir: Optional[Union[Path, str]] = None,
    **kwargs,
) -> Generator[str, None, None]:
    """
    Context manager to create a temporary directory and safely delete it.

    If tmp directory cannot be deleted normally, we set the WRITE permission and retry.
    If cleanup still fails, we give up but don't raise an exception. This is equivalent
    to  `tempfile.TemporaryDirectory(..., ignore_cleanup_errors=True)` introduced in
    Python 3.10.

    See https://www.scivision.dev/python-tempfile-permission-error-windows/.
    """
    tmpdir = tempfile.TemporaryDirectory(
        prefix=prefix, suffix=suffix, dir=dir, **kwargs
    )
    yield tmpdir.name

    try:
        # First once with normal cleanup
        shutil.rmtree(tmpdir.name)
    except Exception:
        # If failed, try to set write permission and retry
        try:
            shutil.rmtree(tmpdir.name, onerror=_set_write_permission_and_retry)
        except Exception:
            pass

    # And finally, cleanup the tmpdir.
    # If it fails again, give up but do not throw error
    try:
        tmpdir.cleanup()
    except Exception:
        pass


# vendored from distutils.util
def strtobool(val):
    """Convert a string representation of truth to true (1) or false (0).

    True values are 'y', 'yes', 't', 'true', 'on', and '1'; false values are 'n', 'no', 'f', 'false', 'off', and '0'.
    Raises ValueError if 'val' is anything else.
    """
    val = val.lower()
    if val in {"y", "yes", "t", "true", "on", "1"}:
        return 1
    if val in {"n", "no", "f", "false", "off", "0"}:
        return 0
    raise ValueError(f"invalid truth value {val!r}")


def infer_framework_from_repr(x):
    """
    Tries to guess the framework of an object `x` from its repr (brittle but will help in `is_tensor` to try the
    frameworks in a smart order, without the need to import the frameworks).
    """
    representation = str(type(x))
    if representation.startswith("<class 'torch."):
        return "pt"
    elif representation.startswith("<class 'tensorflow."):
        return "tf"
    elif representation.startswith("<class 'jax"):
        return "jax"
    elif representation.startswith("<class 'numpy."):
        return "np"


def _get_frameworks_and_test_func(x):
    """
    Returns an (ordered since we are in Python 3.7+) dictionary framework to test function, which places the framework
    we can guess from the repr first, then Numpy, then the others.
    """
    framework_to_test = {
        "np": is_numpy_array,
        "pl": is_polars_dataframe,
    }
    preferred_framework = infer_framework_from_repr(x)
    # We will test this one first, then numpy, then the others.
    frameworks = [] if preferred_framework is None else [preferred_framework]
    if preferred_framework != "np":
        frameworks.append("np")
    frameworks.extend(
        [f for f in framework_to_test if f not in [preferred_framework, "np"]]
    )
    return {f: framework_to_test[f] for f in frameworks}


def is_tensor(x):
    """
    Tests if `x` is a `torch.Tensor`, `tf.Tensor`, `jaxlib.xla_extension.DeviceArray` or `np.ndarray` in the order
    defined by `infer_framework_from_repr`
    """
    # This gives us a smart order to test the frameworks with the corresponding tests.
    framework_to_test_func = _get_frameworks_and_test_func(x)
    for test_func in framework_to_test_func.values():
        if test_func(x):
            return True

    return False


def _is_numpy(x):
    return isinstance(x, np.ndarray)


def _is_polars(x):
    if pl is None:
        return False
    return isinstance(x, pl.DataFrame)


def is_numpy_array(x):
    """
    Tests if `x` is a numpy array or not.
    """
    return _is_numpy(x)


def is_polars_dataframe(x):
    """
    Tests if `x` is a polars dataframe or not.
    """
    return _is_polars(x)


def to_py_obj(obj):
    """
    Convert a TensorFlow tensor, PyTorch tensor, Numpy array or python list to a python list.
    """

    framework_to_py_obj = {
        "pt": lambda obj: obj.detach().cpu().tolist(),
        "tf": lambda obj: obj.numpy().tolist(),
        "jax": lambda obj: np.asarray(obj).tolist(),
        "np": lambda obj: obj.tolist(),
    }

    if isinstance(obj, (dict, UserDict)):
        return {k: to_py_obj(v) for k, v in obj.items()}
    elif isinstance(obj, (list, tuple)):
        return [to_py_obj(o) for o in obj]

    # This gives us a smart order to test the frameworks with the corresponding tests.
    framework_to_test_func = _get_frameworks_and_test_func(obj)
    for framework, test_func in framework_to_test_func.items():
        if test_func(obj):
            return framework_to_py_obj[framework](obj)

    # tolist also works on 0d np arrays
    if isinstance(obj, np.number):
        return obj.tolist()
    else:
        return obj


def to_numpy(obj):
    """
    Convert a TensorFlow tensor, PyTorch tensor, Numpy array or python list to a Numpy array.
    """

    framework_to_numpy = {
        "np": lambda obj: obj,
    }

    if isinstance(obj, (dict, UserDict)):
        return {k: to_numpy(v) for k, v in obj.items()}
    elif isinstance(obj, (list, tuple)):
        return np.array(obj)

    # This gives us a smart order to test the frameworks with the corresponding tests.
    framework_to_test_func = _get_frameworks_and_test_func(obj)
    for framework, test_func in framework_to_test_func.items():
        if test_func(obj):
            return framework_to_numpy[framework](obj)

    return obj


class ModelOutput(OrderedDict):
    """
    Base class for all model outputs as dataclass. Has a `__getitem__` that allows indexing by integer or slice (like a
    tuple) or strings (like a dictionary) that will ignore the `None` attributes. Otherwise behaves like a regular
    python dictionary.

    <Tip warning={true}>

    You can't unpack a `ModelOutput` directly. Use the [`~utils.ModelOutput.to_tuple`] method to convert it to a tuple
    before.

    </Tip>
    """

    def __init_subclass__(cls) -> None:
        """Register subclasses as pytree nodes.

        This is necessary to synchronize gradients when using `torch.nn.parallel.DistributedDataParallel` with
        `static_graph=True` with modules that output `ModelOutput` subclasses.
        """
        pass

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        # Subclasses of ModelOutput must use the @dataclass decorator
        # This check is done in __init__ because the @dataclass decorator operates after __init_subclass__
        # issubclass() would return True for issubclass(ModelOutput, ModelOutput) when False is needed
        # Just need to check that the current class is not ModelOutput
        is_modeloutput_subclass = self.__class__ != ModelOutput

        if is_modeloutput_subclass and not is_dataclass(self):
            raise TypeError(
                f"{self.__module__}.{self.__class__.__name__} is not a dataclasss."
                " This is a subclass of ModelOutput and so must use the @dataclass decorator."
            )

    def __post_init__(self):
        """Check the ModelOutput dataclass.

        Only occurs if @dataclass decorator has been used.
        """
        class_fields = fields(self)

        # Safety and consistency checks
        if not len(class_fields):
            raise ValueError(f"{self.__class__.__name__} has no fields.")
        if not all(field.default is None for field in class_fields[1:]):
            raise ValueError(
                f"{self.__class__.__name__} should not have more than one required field."
            )

        first_field = getattr(self, class_fields[0].name)
        other_fields_are_none = all(
            getattr(self, field.name) is None for field in class_fields[1:]
        )

        if other_fields_are_none and not is_tensor(first_field):
            if isinstance(first_field, dict):
                iterator = first_field.items()
                first_field_iterator = True
            else:
                try:
                    iterator = iter(first_field)
                    first_field_iterator = True
                except TypeError:
                    first_field_iterator = False

            # if we provided an iterator as first field and the iterator is a (key, value) iterator
            # set the associated fields
            if first_field_iterator:
                for idx, element in enumerate(iterator):
                    if (
                        not isinstance(element, (list, tuple))
                        or not len(element) == 2
                        or not isinstance(element[0], str)
                    ):
                        if idx == 0:
                            # If we do not have an iterator of key/values, set it as attribute
                            self[class_fields[0].name] = first_field
                        else:
                            # If we have a mixed iterator, raise an error
                            raise ValueError(
                                f"Cannot set key/value for {element}. It needs to be a tuple (key, value)."
                            )
                        break
                    setattr(self, element[0], element[1])
                    if element[1] is not None:
                        self[element[0]] = element[1]
            elif first_field is not None:
                self[class_fields[0].name] = first_field
        else:
            for field in class_fields:
                v = getattr(self, field.name)
                if v is not None:
                    self[field.name] = v

    def __delitem__(self, *args, **kwargs):
        raise Exception(
            f"You cannot use ``__delitem__`` on a {self.__class__.__name__} instance."
        )

    def setdefault(self, *args, **kwargs):
        raise Exception(
            f"You cannot use ``setdefault`` on a {self.__class__.__name__} instance."
        )

    def pop(self, *args, **kwargs):
        raise Exception(
            f"You cannot use ``pop`` on a {self.__class__.__name__} instance."
        )

    def update(self, *args, **kwargs):
        raise Exception(
            f"You cannot use ``update`` on a {self.__class__.__name__} instance."
        )

    def __getitem__(self, k):
        if isinstance(k, str):
            inner_dict = dict(self.items())
            return inner_dict[k]
        else:
            return self.to_tuple()[k]

    def __setattr__(self, name, value):
        if name in self.keys() and value is not None:
            # Don't call self.__setitem__ to avoid recursion errors
            super().__setitem__(name, value)
        super().__setattr__(name, value)

    def __setitem__(self, key, value):
        # Will raise a KeyException if needed
        super().__setitem__(key, value)
        # Don't call self.__setattr__ to avoid recursion errors
        super().__setattr__(key, value)

    def __reduce__(self):
        if not is_dataclass(self):
            return super().__reduce__()
        callable, _args, *remaining = super().__reduce__()
        args = tuple(getattr(self, field.name) for field in fields(self))
        return callable, args, *remaining

    def to_tuple(self) -> Tuple[Any]:
        """
        Convert self to a tuple containing all the attributes/keys that are not `None`.
        """
        return tuple(self[k] for k in self.keys())


class ExplicitEnum(str, Enum):
    """
    Enum with more explicit error message for missing values.
    """

    @classmethod
    def _missing_(cls, value):
        raise ValueError(
            f"{value} is not a valid {cls.__name__}, please select one of {list(cls._value2member_map_.keys())}"
        )


class ContextManagers:
    """
    Wrapper for `contextlib.ExitStack` which enters a collection of context managers. Adaptation of `ContextManagers`
    in the `fastcore` library.
    """

    def __init__(self, context_managers: List[ContextManager]):
        self.context_managers = context_managers
        self.stack = ExitStack()

    def __enter__(self):
        for context_manager in self.context_managers:
            self.stack.enter_context(context_manager)

    def __exit__(self, *args, **kwargs):
        self.stack.__exit__(*args, **kwargs)


def can_return_loss(model_class):
    """
    Check if a given model can return loss.

    Args:
        model_class (`type`): The class of the model.
    """
    framework = infer_framework(model_class)
    if framework == "tf":
        signature = inspect.signature(model_class.call)  # TensorFlow models
    elif framework == "pt":
        signature = inspect.signature(model_class.forward)  # PyTorch models
    else:
        signature = inspect.signature(model_class.__call__)  # Flax models

    for p in signature.parameters:
        if p == "return_loss" and signature.parameters[p].default is True:
            return True

    return False


def find_labels(model_class):
    """
    Find the labels used by a given model.

    Args:
        model_class (`type`): The class of the model.
    """
    model_name = model_class.__name__
    infer_framework(model_class)
    signature = inspect.signature(model_class.__call__)  # Flax models

    if "QuestionAnswering" in model_name:
        return [
            p
            for p in signature.parameters
            if "label" in p or p in ("start_positions", "end_positions")
        ]
    else:
        return [p for p in signature.parameters if "label" in p]


def flatten_dict(d: MutableMapping, parent_key: str = "", delimiter: str = "."):
    """Flatten a nested dict into a single level dict."""

    def _flatten_dict(d, parent_key="", delimiter="."):
        for k, v in d.items():
            key = str(parent_key) + delimiter + str(k) if parent_key else k
            if v and isinstance(v, MutableMapping):
                yield from flatten_dict(v, key, delimiter=delimiter).items()
            else:
                yield key, v

    return dict(_flatten_dict(d, parent_key, delimiter))


@contextmanager
def working_or_temp_dir(working_dir, use_temp_dir: bool = False):
    if use_temp_dir:
        with tempfile.TemporaryDirectory() as tmp_dir:
            yield tmp_dir
    else:
        yield working_dir


def add_model_info_to_auto_map(auto_map, repo_id):
    """
    Adds the information of the repo_id to a given auto map.
    """
    for key, value in auto_map.items():
        if isinstance(value, (tuple, list)):
            auto_map[key] = [
                f"{repo_id}--{v}" if (v is not None and "--" not in v) else v
                for v in value
            ]
        elif value is not None and "--" not in value:
            auto_map[key] = f"{repo_id}--{value}"

    return auto_map


def infer_framework(model_class):
    """
    Infers the framework of a given model without using isinstance(), because we cannot guarantee that the relevant
    classes are imported or available.
    """
    for base_class in inspect.getmro(model_class):
        module = base_class.__module__
        name = base_class.__name__
        if (
            module.startswith("tensorflow")
            or module.startswith("keras")
            or name == "TFPreTrainedModel"
        ):
            return "tf"
        elif module.startswith("torch") or name == "PreTrainedModel":
            return "pt"
        elif (
            module.startswith("flax")
            or module.startswith("jax")
            or name == "FlaxPreTrainedModel"
        ):
            return "flax"
    else:
        raise TypeError(f"Could not infer framework from class {model_class}.")



"""
This module provides classes for computing the sum of rows and columns in input data.

Classes:
- RowSumStat: Computes the sum of each row in the input data.
- ColumnSumStat: Computes the sum of each column in the input data.
- ColumnStatForOTU: Computes the sum of each column in the OTUAbundance feature.
- RowSumStatForOTU: Computes the sum of each row in the OTUAbundance feature.
- TotalOTUAbundanceStat: Alias for RowSumStatForOTU.

These classes provide methods for transforming and fitting the input data using different libraries such as numpy, pandas, and polars.

Note: The ColumnStatForOTU and RowSumStatForOTU classes are provided for autostat and are equivalent to ColumnSumStat and RowSumStat respectively.
"""

from dataclasses import dataclass, field
from typing import TYPE_CHECKING, List, Type

import numpy as np
import pandas as pd
from biocore.utils.import_util import requires_backends

from biofit.integration.biosets import get_feature
from biofit.processing import SelectedColumnTypes
from biofit.utils import logging

from ..stat import Stat, StatConfig

if TYPE_CHECKING:
    import polars as pl

logger = logging.get_logger(__name__)


@dataclass
class RowSumStatConfig(StatConfig):
    # process description
    _transform_process_desc: str = field(
        default="Calculating row sums", init=False, repr=False
    )
    _fit_unused_feature_types: List[Type] = field(
        default_factory=lambda: [get_feature("METADATA_FEATURE_TYPES")],
        init=False,
        repr=False,
    )
    _transform_unused_feature_types: List[Type] = field(
        default_factory=lambda: [get_feature("METADATA_FEATURE_TYPES")],
        init=False,
        repr=False,
    )
    processor_name: str = field(default="row_sum", init=False, repr=False)
    _n_features_out: int = field(default=1, init=False, repr=False)


@dataclass
class RowSumStatConfigForOTU(RowSumStatConfig):
    """Computes the sum of each row in the OTUAbundance feature."""

    # dataset specific attributes
    _fit_input_feature_types: List[Type] = field(
        default_factory=lambda: [get_feature("Abundance")], init=False, repr=False
    )
    _transform_input_feature_types: List[Type] = field(
        default_factory=lambda: [get_feature("Abundance")], init=False, repr=False
    )
    dataset_name: str = field(default="otu", init=False, repr=False)


class RowSumStat(Stat):
    """
    Computes the sum of each row in the input data.
    """

    # config attributes
    _config_class = RowSumStatConfig
    config: RowSumStatConfig
    output_dtype = "float64"

    def fit(
        self,
        X,
        input_columns: SelectedColumnTypes = None,
        raise_if_missing: bool = None,
        cache_output: bool = None,
        load_from_cache_file: bool = None,
        batched: bool = True,
        batch_size: int = 1000,
        batch_format: str = None,
        num_proc: int = None,
        map_kwargs: dict = {"fn_kwargs": {}},
        cache_dir: str = None,
        cache_file_name: str = None,
        fingerprint: str = None,
    ) -> "RowSumStat":
        self.config._input_columns = self._set_input_columns_and_arity(input_columns)
        return self._process_fit(
            X,
            raise_if_missing=raise_if_missing,
            cache_output=cache_output,
            load_from_cache_file=load_from_cache_file,
            batched=batched,
            batch_size=batch_size,
            batch_format=batch_format,
            num_proc=num_proc,
            map_kwargs=map_kwargs,
            cache_dir=cache_dir,
            cache_file_name=cache_file_name,
            fingerprint=fingerprint,
        )

    def _transform_numpy(self, X: np.ndarray):
        """
        Transforms the input data using numpy.

        Args:
            X: Input data as a numpy array.

        Returns:
            Transformed data as a numpy array.
        """
        if len(X.shape) == 1:
            return np.sum(X)[:, None]
        return np.sum(X, axis=1)[:, None]

    def _transform_pandas(self, X: pd.DataFrame):
        """
        Transforms the input data using pandas.

        Args:
            X: Input data as a pandas DataFrame.

        Returns:
            Transformed data as a pandas DataFrame.
        """
        return X.astype("float64").sum(axis=1).to_frame()

    def _transform_polars(self, X: "pl.DataFrame"):
        """
        Transforms the input data using polars.

        Args:
            X: Input data as a polars DataFrame.

        Returns:
            Transformed data as a polars DataFrame.
        """
        requires_backends(self._transform_polars, "polars")
        import polars as pl

        return X.cast(pl.Float64).sum_horizontal().to_frame()



import importlib

from biocore.utils.import_util import is_biosets_available


class NotAvailable:
    def __init__(self, *args, **kwargs):
        pass


def get_feature(val):
    if is_biosets_available():
        return getattr(importlib.import_module("biosets.features"), val)
    return None



from dataclasses import dataclass, field

from biofit.processing import BaseProcessor, ProcessorConfig


@dataclass
class EncoderConfig(ProcessorConfig):
    """Base class for feature extraction processor configurations."""

    processor_type: str = field(default="encoding", init=False, repr=False)


class Encoder(BaseProcessor):
    """Base class for feature extraction processors."""



from typing import List, Union

from biocore.utils.import_util import is_biosets_available
from biocore.utils.inspect import get_kwargs

from biofit.auto.auto_factory import (
    _BaseAutoProcessorClass,
    _LazyAutoMapping,
)
from biofit.processing import BaseProcessor
from biofit.visualization.plotting import BasePlotter
from biofit.visualization.plotting_utils import (
    display_image_carousel,
    is_in_notebook,
)

from .configuration_auto import (
    DATASET_PLT_TO_MAPPER_NAMES,
    PLOTTER_CONFIG_MAPPING_NAMES,
    PLOTTER_MAPPING_NAMES,
    AutoPlotterConfig,
)
from .processing_auto import AutoPreprocessor, ProcessorPipeline

PLOTTER_MAPPING = _LazyAutoMapping(PLOTTER_CONFIG_MAPPING_NAMES, PLOTTER_MAPPING_NAMES)


class PlotterPipeline:
    def __init__(self, plotters: List[BasePlotter], processors: List[BaseProcessor]):
        self.plotters = plotters
        self.processors = processors

    def plot(self, X, *args, fit=True, **kwargs):
        from datasets import Dataset, IterableDataset

        from biofit import Bioset

        if not isinstance(X, (Bioset, Dataset, IterableDataset)):
            raise ValueError("X must be a Bioset or huggingface Dataset.")
        pre_X = X
        show = kwargs.pop("show", True)
        images = []
        for plotter, processor in zip(self.plotters, self.processors):
            fit_trans_kwargs = get_kwargs(kwargs, processor.fit_transform)
            if fit:
                after_X = processor.fit_transform(pre_X, *args, **fit_trans_kwargs)
            else:
                after_X = processor.transform(pre_X, *args, **fit_trans_kwargs)
            if plotter.config._compare:
                path = plotter.plot(pre_X, after_X, *args, show=False, **kwargs)
            else:
                path = plotter.plot(after_X, *args, show=False, **kwargs)
            if isinstance(path, list):
                images.extend(path)
            else:
                images.append(path)
            pre_X = after_X

        if show and is_in_notebook():
            display_image_carousel(images)


class AutoPlotter(_BaseAutoProcessorClass):
    _processor_mapping = PLOTTER_MAPPING

    @classmethod
    def for_dataset(cls, dataset_name, **kwargs):
        """Create a processor for a dataset.

        Args:
            dataset (Bioset): The dataset to create a processor for.

        Returns:
            Processor: The processor for the dataset.
        """

        if is_biosets_available():
            from biosets.packaged_modules import EXPERIMENT_TYPE_ALIAS
        else:
            EXPERIMENT_TYPE_ALIAS = {}

        dataset_name = EXPERIMENT_TYPE_ALIAS.get(dataset_name, dataset_name)
        _plotter_mapping = _LazyAutoMapping(
            DATASET_PLT_TO_MAPPER_NAMES.get(dataset_name), PLOTTER_MAPPING_NAMES
        )
        configs = AutoPlotterConfig.for_dataset(dataset_name)
        procs = []
        for config in configs:
            config_kwargs = get_kwargs(kwargs, config.__class__.__init__)
            procs.append(
                _plotter_mapping[type(config)]._from_config(config, **config_kwargs)
            )

        processors = AutoPreprocessor.for_dataset(dataset_name)

        return PlotterPipeline(procs, processors)

    @classmethod
    def from_processor(
        cls, processor, dataset_name=None, **kwargs
    ) -> Union[PlotterPipeline, BasePlotter]:
        """Create a processor from another processor.

        Args:
            processor (Processor): The processor to create a processor for.

        Returns:
            Processor: The processor for the dataset.
        """

        def get_proc(proc, dataset_name=None, **kwargs):
            dataset_name = dataset_name or proc.config.dataset_name
            if dataset_name:
                if is_biosets_available():
                    from biosets.packaged_modules import EXPERIMENT_TYPE_ALIAS
                else:
                    EXPERIMENT_TYPE_ALIAS = {}
                dataset_name = EXPERIMENT_TYPE_ALIAS.get(dataset_name, dataset_name)
                _plotter_mapping = _LazyAutoMapping(
                    DATASET_PLT_TO_MAPPER_NAMES.get(dataset_name),
                    PLOTTER_MAPPING_NAMES,
                )
                config = AutoPlotterConfig.for_processor(
                    proc.config.processor_name,
                    dataset_name=dataset_name,
                )
            else:
                _plotter_mapping = PLOTTER_MAPPING
                config = AutoPlotterConfig.for_processor(proc.config.processor_name)
            config_kwargs = get_kwargs(kwargs, config.__class__.__init__)
            return _plotter_mapping[type(config)]._from_config(config, **config_kwargs)

        if isinstance(processor, ProcessorPipeline):
            plotters = []
            for proc in processor.steps:
                plotters.append(get_proc(proc[1], dataset_name=dataset_name, **kwargs))
            return PlotterPipeline(plotters, processor.processors)
        elif isinstance(processor, BaseProcessor):
            return get_proc(processor, **kwargs)
        else:
            raise ValueError(
                "processor must be a `biofit.ProcessorPipeline` or `biofit.BaseProcessor`."
            )



import sys
from os import PathLike
from typing import List, Optional, Union

import numpy as np
import pandas as pd
from biocore import DataHandler
from biocore.utils import requires_backends
from sklearn.pipeline import Pipeline

from biofit.processing import SelectedColumnTypes
from biofit.utils.types import Unset


def is_in_notebook():
    try:
        from IPython import get_ipython

        if "IPKernelApp" in get_ipython().config:
            return True
    except Exception:
        return False
    return False


def display_image_carousel(image_paths, format="png"):
    """
    Displays an image carousel with left and right arrow buttons in a Jupyter notebook.

    Parameters:
        image_paths (list): List of paths to images.
        output_dir (str): Directory to indicate in the log if limit is exceeded.
        max_images (int): Maximum number of images to include in the carousel.
    """
    requires_backends(display_image_carousel, "ipywidgets")
    import ipywidgets as widgets
    from IPython.display import Image, clear_output, display
    from ipywidgets import Button, HBox, Layout, VBox

    image_index = 0

    button_layout = Layout(width="100px")
    left_button = Button(description="Previous", layout=button_layout)
    right_button = Button(description="Next", layout=button_layout)
    image_box = widgets.Output()

    def show_image():
        """Helper function to show the image."""
        with image_box:
            clear_output(wait=True)
            display(
                Image(image_paths[image_index], embed=True, format=format, width=720)
            )

    def on_left_button_clicked(b):
        nonlocal image_index
        if image_index > 0:
            image_index -= 1
            show_image()

    def on_right_button_clicked(b):
        nonlocal image_index
        if image_index < len(image_paths) - 1:
            image_index += 1
            show_image()

    left_button.on_click(on_left_button_clicked)
    right_button.on_click(on_right_button_clicked)

    show_image()  # Show the first image initially

    navigation_box = HBox([left_button, right_button])
    display(VBox([navigation_box, image_box]))


def get_distinct_colors(n, colormap="nipy_spectral"):
    """
    Generate a list of n distinct colors using a specified colormap.
    Switched to 'nipy_spectral' for higher contrast and brightness differentiation.

    :param n: Number of distinct colors to generate.
    :param colormap: Name of the matplotlib colormap to use.
    :return: List of RGBA colors.
    """
    requires_backends(get_distinct_colors, "matplotlib")
    from matplotlib import colormaps

    cmap = colormaps.get_cmap(colormap)
    return [cmap(i / n) for i in range(n)]


def generate_violin(
    x,
    y=None,
    column: SelectedColumnTypes = None,
    label_name: SelectedColumnTypes = None,
    xlab: str = Unset('"Labels"'),
    ylab: str = Unset('"Value"'),
    device: str = "pdf",
    fingerprint: str = None,
    unused_columns: SelectedColumnTypes = None,
    raise_if_missing: bool = True,
    output_dir: Union[PathLike, str] = None,
):
    """Generates a violin plot of the input data.

    Args:
        x: Input data.
        other_x: Other input data.
        y: Target data.
        other_x: Other target data.
        **kwargs: Additional arguments.
    Returns:
        Plotter object.
    """
    from biofit.visualization.violin import ViolinPlotter

    return ViolinPlotter().plot(
        x,
        y,
        column=column,
        label_name=label_name,
        xlab=xlab,
        ylab=ylab,
        path=output_dir,
        device=device,
        fingerprint=fingerprint,
        unused_columns=unused_columns,
        raise_if_missing=raise_if_missing,
    )


def generate_scatterplot(
    x,
    y=None,
    group=None,
    xdata: SelectedColumnTypes = None,
    ydata: SelectedColumnTypes = None,
    groupby: SelectedColumnTypes = None,
    xlab: str = Unset("None"),
    ylab: str = Unset("None"),
    title: str = Unset('"Scatterplot"'),
    alpha: str = Unset("1"),
    col_set: str = Unset('"Set1"'),
    cols: List[str] = Unset("None"),
    xlog: str = Unset("None"),
    ylog: str = Unset("None"),
    output_dir: str = None,
    device: str = "pdf",
    fingerprint: str = None,
    unused_columns: SelectedColumnTypes = None,
    raise_if_missing: bool = True,
):
    from biofit.visualization.scatterplot import ScatterPlotter

    plotter = ScatterPlotter()
    return plotter.plot(
        x,
        y,
        group,
        xdata=xdata,
        ydata=ydata,
        groupby=groupby,
        xlab=xlab,
        ylab=ylab,
        title=title,
        alpha=alpha,
        col_set=col_set,
        cols=cols,
        xlog=xlog,
        ylog=ylog,
        path=output_dir,
        device=device,
        fingerprint=fingerprint,
        unused_columns=unused_columns,
        raise_if_missing=raise_if_missing,
    )


def generate_histogram(
    x,
    input_columns: SelectedColumnTypes = None,
    xlab: str = Unset('"X"'),
    ylab: str = Unset('"Frequency"'),
    title: str = Unset('"Histogram"'),
    bins: int = Unset("30"),
    font_size: int = Unset("8"),
    col_fill: str = Unset('"grey40"'),
    col_outline: str = Unset('"white"'),
    xlog: Optional[str] = Unset("None"),
    ylog: Optional[str] = Unset("None"),
    output_dir: str = None,
    device: str = "pdf",
    fingerprint: str = None,
    unused_columns: SelectedColumnTypes = None,
    raise_if_missing: bool = True,
):
    """Generates a histogram of the input data.

    Args:
        x: Input data.
        other_x: Other input data.
        y: Target data.
        other_x: Other target data.
        **kwargs: Additional arguments.
    Returns:
        Plotter object.
    """
    from biofit.visualization.histogram import HistogramPlotter

    plotter = HistogramPlotter()

    return plotter.plot(
        x,
        input_columns=input_columns,
        xlab=xlab,
        ylab=ylab,
        title=title,
        bins=bins,
        font_size=font_size,
        col_fill=col_fill,
        col_outline=col_outline,
        xlog=xlog,
        ylog=ylog,
        path=output_dir,
        device=device,
        fingerprint=fingerprint,
        unused_columns=unused_columns,
        raise_if_missing=raise_if_missing,
    )


def generate_barplot(
    x,
    y=None,
    group=None,
    label_name: SelectedColumnTypes = None,
    value_name=None,
    groupby: Optional[str] = None,
    xlab: Optional[str] = Unset("None"),
    ylab: Optional[str] = Unset("None"),
    title: str = Unset('"Bar Plot"'),
    col_set: str = Unset('"Set1"'),
    col_labels: str = Unset('"black"'),
    col_outline: str = Unset('"grey30"'),
    cols: Optional[List[str]] = Unset("None"),
    prop: bool = Unset("False"),
    add_count_lab: bool = Unset("True"),
    vars_as_entered: bool = Unset("False"),
    legend_position: str = Unset('"top"'),
    font_size: float = Unset("3.25"),
    output_dir: str = None,
    device: str = "pdf",
    fingerprint: str = None,
    unused_columns: SelectedColumnTypes = None,
    raise_if_missing: bool = True,
):
    """Generates a bar plot of the input data.

    Args:
        x: Input data.
        other_x: Other input data.
        y: Target data.
        other_x: Other target data.
        **kwargs: Additional arguments.
    Returns:
        Plotter object.
    """
    from biofit.visualization.barplot import BarPlotter

    plotter = BarPlotter()

    return plotter.plot(
        x,
        y,
        group,
        label_name=label_name,
        value_name=value_name,
        groupby=groupby,
        xlab=xlab,
        ylab=ylab,
        title=title,
        col_set=col_set,
        col_labels=col_labels,
        col_outline=col_outline,
        cols=cols,
        prop=prop,
        add_count_lab=add_count_lab,
        vars_as_entered=vars_as_entered,
        legend_position=legend_position,
        font_size=font_size,
        path=output_dir,
        device=device,
        fingerprint=fingerprint,
        unused_columns=unused_columns,
        raise_if_missing=raise_if_missing,
    )


def generate_comparison_histogram(
    x1,
    x2=None,
    column1: str = None,
    column2: str = None,
    xlab: Optional[str] = Unset("None"),
    ylab: str = Unset("None"),
    title: str = Unset("None"),
    bins: int = Unset("None"),
    alpha: float = Unset("None"),
    legend_title: str = Unset("None"),
    legend_position: str = Unset("None"),
    subplot_title1: str = Unset("None"),
    subplot_title2: str = Unset("None"),
    col_set: str = Unset("None"),
    cols: Optional[List[str]] = Unset("None"),
    xlog: Optional[bool] = Unset("None"),
    ylog: Optional[bool] = Unset("None"),
    device: str = "pdf",
    fingerprint: str = None,
    unused_columns: SelectedColumnTypes = None,
    raise_if_missing: bool = True,
    output_dir=None,
):
    """Generates a comparison histogram of the input data.

    Args:
        x: Input data.
        other_x: Other input data.
        y: Target data.
        other_x: Other target data.
        **kwargs: Additional arguments.
    Returns:
        Plotter object.
    """
    from biofit.visualization.histogram import ComparisonHistogramPlotter

    plotter = ComparisonHistogramPlotter()

    return plotter.plot(
        x1,
        x2=x2,
        column1=column1,
        column2=column2,
        xlab=xlab,
        ylab=ylab,
        title=title,
        bins=bins,
        alpha=alpha,
        legend_title=legend_title,
        legend_position=legend_position,
        col_set=col_set,
        cols=cols,
        subplot_title1=subplot_title1,
        subplot_title2=subplot_title2,
        xlog=xlog,
        ylog=ylog,
        path=output_dir,
        device=device,
        fingerprint=fingerprint,
        unused_columns=unused_columns,
        raise_if_missing=raise_if_missing,
    )


def plot_correlation(
    X,
    y=None,
    group=None,
    input_columns=None,
    target_column=None,
    groupby: Optional[str] = None,
    precomputed=False,
    method="auto",
    label_name: str = Unset('"None"'),
    value_name: Optional[str] = Unset("None"),
    top_k: int = Unset("30"),
    xlab: Optional[str] = Unset("None"),
    ylab: Optional[str] = Unset("None"),
    title: str = Unset('"Bar Plot"'),
    col_set: str = Unset('"Set1"'),
    cols: Optional[List[str]] = Unset("None"),
    prop: bool = Unset("False"),
    add_count_lab: bool = Unset("True"),
    vars_as_entered: bool = Unset("False"),
    legend_position: str = Unset('"top"'),
    font_size: float = Unset("3.25"),
    output_dir=None,
    file_name=None,
):
    """Plot the correlation matrix of the input data.

    Args:
        X: Input data
        y: Target variable
        input_columns: Columns to filter
        target_column: Target column
        top_k: Number of top features to plot
        label_name: Name of the labels
        value_name: Name of the values
        groupby: Groupby column
        xlab: X-axis label
        ylab: Y-axis label
        title: Plot title
        col_set: Color set
        cols: Columns to plot
        prop: Whether to plot proportions
        add_count_lab: Whether to add count labels
        vars_as_entered: Whether the variables are entered as is
        legend_position: Position of the legend
        font_size: Font size
    Returns:
        SampleFiltered data.
    """
    if not precomputed:
        from biofit.stat import CorrelationStat

        corr_stat = CorrelationStat(method=method)
        corrs = corr_stat.fit_transform(
            X, y, input_columns, target_column, output_format="pandas"
        )
        name_map = {
            "pearsonr": "Pearson Correlation",
            "spearmanr": "Spearman Correlation",
            "kendalltau": "Kendall Correlation",
            "pointbiserialr": "Point Biserial Correlation",
        }
        value_name = name_map.get(corr_stat.config.method, "Correlation")
        if groupby is None or isinstance(groupby, Unset):
            groupby = "Features"
        corrs = corrs.melt(var_name=groupby, value_name=value_name)
        # drop na
        corrs = corrs.dropna()
        sorted_inds = np.argsort(np.abs(corrs[value_name]))[::-1]
        if groupby is None or isinstance(top_k, Unset):
            top_k = 30
        corrs = corrs.iloc[sorted_inds[:top_k]]
        return generate_barplot(
            corrs,
            None,
            None,
            label_name=label_name,
            value_name=value_name,
            groupby=groupby,
            xlab=xlab,
            ylab=ylab,
            title=title,
            col_set=col_set,
            cols=cols,
            prop=prop,
            add_count_lab=add_count_lab,
            vars_as_entered=vars_as_entered,
            legend_position=legend_position,
            font_size=font_size,
            output_dir=output_dir,
        )

    return generate_barplot(
        X,
        y,
        None,
        label_name=label_name,
        value_name=value_name,
        groupby=groupby,
        xlab=xlab,
        ylab=ylab,
        title=title,
        col_set=col_set,
        cols=cols,
        prop=prop,
        add_count_lab=add_count_lab,
        vars_as_entered=vars_as_entered,
        legend_position=legend_position,
        font_size=font_size,
        output_dir=output_dir,
    )


def plot_feature_distribution(
    X,
    input_columns: SelectedColumnTypes = None,
    value_name=None,
    aggregate=None,
    aggregate_kwargs={},
    xlab: str = Unset('"X"'),
    ylab: str = Unset('"Frequency"'),
    title: str = Unset('"Histogram"'),
    bins: int = Unset("30"),
    font_size: int = Unset("8"),
    col_fill: str = Unset('"grey40"'),
    col_outline: str = Unset('"white"'),
    xlog: Optional[str] = Unset("None"),
    ylog: Optional[str] = Unset("None"),
    output_dir: str = None,
    device: str = "pdf",
    fingerprint: str = None,
    unused_columns: SelectedColumnTypes = None,
    raise_if_missing: bool = True,
):
    """Plot the feature distribution of the input data.

    Args:
        X: Input data
        columns: Columns to plot
        aggregate: Aggregate each feature by: 'mean', 'median', 'sum', 'std', 'var', 'min', 'max', or 'presence'
        **kwargs: Additional keyword arguments

    Returns:
        Plot object.
    """
    x_dims = DataHandler.get_shape(X)
    precomputed = aggregate is None and (
        value_name is not None or len(x_dims) == 1 or x_dims[1] == 1
    )
    value_name = value_name or "Presence"

    if aggregate == "presence":
        from biofit.stat import ColumnMissingnessStat

        num_rows = DataHandler.get_shape(X)[0]
        missingness = ColumnMissingnessStat(
            input_columns=input_columns, **aggregate_kwargs
        ).fit_transform(X, output_format="pandas")
        data = num_rows - missingness
    elif aggregate == "sum":
        from biofit.stat import ColumnSumStat

        data = ColumnSumStat(
            input_columns=input_columns, **aggregate_kwargs
        ).fit_transform(X, output_format="pandas")
    elif aggregate == "mean":
        from biofit.stat import ColumnMeanStat

        data = ColumnMeanStat(
            input_columns=input_columns, **aggregate_kwargs
        ).fit_transform(X, output_format="pandas")

    if (
        input_columns is None
        and "biosets" in sys.modules
        and isinstance(X, getattr(sys.modules["biosets"], "Bioset"))
    ):
        from biosets import get_data

        data = get_data(X)
    else:
        data = X

    data = DataHandler.to_pandas(data)

    if input_columns:
        data = data[input_columns]

    if not precomputed:
        data = data.melt(value_name=value_name)

    return generate_histogram(
        data,
        input_columns=input_columns,
        xlab=xlab,
        ylab=ylab,
        title=title,
        bins=bins,
        font_size=font_size,
        col_fill=col_fill,
        col_outline=col_outline,
        xlog=xlog,
        ylog=ylog,
        output_dir=output_dir,
        device=device,
        fingerprint=fingerprint,
        unused_columns=unused_columns,
        raise_if_missing=raise_if_missing,
    )


def compare_feature_distributions(
    X1,
    X2,
    columns1: SelectedColumnTypes = None,
    columns2: SelectedColumnTypes = None,
    value_name=None,
    aggregate=None,
    aggregate_kwargs={},
    xlab: Optional[str] = Unset("None"),
    ylab: str = Unset("None"),
    title: str = Unset("None"),
    bins: int = Unset("None"),
    alpha: float = Unset("None"),
    legend_title: str = Unset("None"),
    legend_position: str = Unset("None"),
    subplot_title1: str = Unset("None"),
    subplot_title2: str = Unset("None"),
    col_set: str = Unset("None"),
    cols: Optional[List[str]] = Unset("None"),
    xlog: Optional[bool] = Unset("None"),
    ylog: Optional[bool] = Unset("None"),
    output_dir: str = None,
    device: str = "pdf",
    fingerprint: str = None,
    unused_columns: SelectedColumnTypes = None,
    raise_if_missing: bool = True,
):
    """Compare the feature distribution of the input data.

    Args:
        X1: Input data 1
        X2: Input data 2
        columns1: Columns to plot for data 1
        columns2: Columns to plot for data 2. If not provided, columns1 will be used for data 2.
        aggregate: Aggregate each feature by: 'mean', 'median', 'sum', 'std', 'var', 'min', 'max', or 'presence'
        **kwargs: Additional keyword arguments
    Returns:
        Plot object.
    """
    x1_dims = DataHandler.get_shape(X1)
    x2_dims = DataHandler.get_shape(X2)
    precomputed = (
        aggregate is None
        and (value_name is not None or len(x1_dims) == 1 or x1_dims[1] == 1)
        and (value_name is not None or len(x2_dims) == 1 or x2_dims[1] == 1)
    )

    if columns1 and not columns2:
        columns2 = columns1

    if aggregate == "presence":
        from biofit.stat import ColumnMissingnessStat

        value_name = value_name or "Presence"
        num_rows1 = DataHandler.get_shape(X1)[0]
        num_rows2 = DataHandler.get_shape(X2)[0]

        missingness1 = ColumnMissingnessStat(
            input_columns=columns1, **aggregate_kwargs
        ).fit_transform(X1, output_format="pandas")
        missingness2 = ColumnMissingnessStat(
            input_columns=columns2, **aggregate_kwargs
        ).fit_transform(X2, output_format="pandas")

        data1 = num_rows1 - missingness1
        data2 = num_rows2 - missingness2
    elif aggregate == "sum":
        from biofit.stat import ColumnSumStat

        value_name = value_name or "Sum"
        data1 = ColumnSumStat(input_columns=columns1, **aggregate_kwargs).fit_transform(
            X1, output_format="pandas"
        )
        data2 = ColumnSumStat(input_columns=columns2, **aggregate_kwargs).fit_transform(
            X2, output_format="pandas"
        )
    elif aggregate == "mean":
        from biofit.stat import ColumnMeanStat

        value_name = value_name or "Mean"
        data1 = ColumnMeanStat(
            input_columns=columns1, **aggregate_kwargs
        ).fit_transform(X1, output_format="pandas")
        data2 = ColumnMeanStat(
            input_columns=columns2, **aggregate_kwargs
        ).fit_transform(X2, output_format="pandas")
    else:
        data1 = X1
        data2 = X2

    value_name = value_name or "Value"

    data1 = DataHandler.to_pandas(data1)
    data2 = DataHandler.to_pandas(data2)

    if columns1:
        data1 = data1[columns1]

    if columns2:
        data2 = data2[columns2]

    if not precomputed:
        if data1.shape[1] > 2:
            data1 = data1.melt(value_name=value_name)

        if data2.shape[1] > 2:
            data2 = data2.melt(value_name=value_name)

    return generate_comparison_histogram(
        data1,
        data2,
        column1=value_name,
        column2=value_name,
        xlab=xlab,
        ylab=ylab,
        title=title,
        bins=bins,
        alpha=alpha,
        legend_title=legend_title,
        legend_position=legend_position,
        col_set=col_set,
        cols=cols,
        subplot_title1=subplot_title1,
        subplot_title2=subplot_title2,
        xlog=xlog,
        ylog=ylog,
        output_dir=output_dir,
        device=device,
        fingerprint=fingerprint,
        unused_columns=unused_columns,
        raise_if_missing=raise_if_missing,
    )


def plot_sample_distribution(
    X,
    input_columns: SelectedColumnTypes = None,
    value_name=None,
    aggregate=None,
    aggregate_kwargs={},
    xlab: str = Unset('"X"'),
    ylab: str = Unset('"Frequency"'),
    title: str = Unset('"Histogram"'),
    bins: int = Unset("30"),
    font_size: int = Unset("8"),
    col_fill: str = Unset('"grey40"'),
    col_outline: str = Unset('"white"'),
    xlog: Optional[str] = Unset("None"),
    ylog: Optional[str] = Unset("None"),
    output_dir: str = None,
    device: str = "pdf",
    fingerprint: str = None,
    unused_columns: SelectedColumnTypes = None,
    raise_if_missing: bool = True,
):
    """Plot the feature distribution of the input data.

    Args:
        X: Input data
        columns: Columns to plot
        aggregate: Aggregate each feature by: 'mean', 'median', 'sum', 'std', 'var', 'min', 'max', or 'presence'
        **kwargs: Additional keyword arguments

    Returns:
        Plot object.
    """
    x_dims = DataHandler.get_shape(X)
    precomputed = aggregate is None and (
        value_name is not None or len(x_dims) == 1 or x_dims[1] == 1
    )

    if aggregate == "presence":
        from biofit.stat import RowMissingnessStat

        value_name = value_name or "Presence"
        num_rows = DataHandler.get_shape(X)[0]
        missingness = RowMissingnessStat(
            input_columns=input_columns, **aggregate_kwargs
        ).fit_transform(X, output_format="pandas")
        data = num_rows - missingness
    elif aggregate == "sum":
        from biofit.stat import RowSumStat

        value_name = value_name or "Sum"
        data = RowSumStat(
            input_columns=input_columns, **aggregate_kwargs
        ).fit_transform(X, output_format="pandas")
    elif aggregate == "mean":
        from biofit.stat import RowMeanStat

        value_name = value_name or "Mean"
        data = RowMeanStat(
            input_columns=input_columns, **aggregate_kwargs
        ).fit_transform(X, output_format="pandas")

    value_name = value_name or "Value"

    if (
        input_columns is None
        and "biosets" in sys.modules
        and isinstance(X, getattr(sys.modules["biosets"], "Bioset"))
    ):
        from biosets import get_data

        data = get_data(X)
    else:
        data = X

    data = DataHandler.to_pandas(data)

    if input_columns:
        data = data[input_columns]

    if not precomputed:
        if data.shape[1] > 2:
            data = data.melt(value_name=value_name)

    return generate_histogram(
        data,
        input_columns=value_name,
        xlab=xlab,
        ylab=ylab,
        title=title,
        bins=bins,
        font_size=font_size,
        col_fill=col_fill,
        col_outline=col_outline,
        xlog=xlog,
        ylog=ylog,
        output_dir=output_dir,
        device=device,
        fingerprint=fingerprint,
        unused_columns=unused_columns,
        raise_if_missing=raise_if_missing,
    )


def compare_sample_distributions(
    X1,
    X2,
    columns1: SelectedColumnTypes = None,
    columns2: SelectedColumnTypes = None,
    value_name=None,
    aggregate=None,
    aggregate_kwargs={},
    xlab: Optional[str] = Unset("None"),
    ylab: str = Unset("None"),
    title: str = Unset("None"),
    bins: int = Unset("None"),
    alpha: float = Unset("None"),
    legend_title: str = Unset("None"),
    legend_position: str = Unset("None"),
    subplot_title1: str = Unset("None"),
    subplot_title2: str = Unset("None"),
    col_set: str = Unset("None"),
    cols: Optional[List[str]] = Unset("None"),
    xlog: Optional[bool] = Unset("None"),
    ylog: Optional[bool] = Unset("None"),
    device: str = "pdf",
    fingerprint: str = None,
    unused_columns: SelectedColumnTypes = None,
    raise_if_missing: bool = True,
    output_dir=None,
):
    """Compare the feature distribution of the input data.

    Args:
        X1: Input data 1
        X2: Input data 2
        columns1: Columns to plot for data 1
        columns2: Columns to plot for data 2. If not provided, columns1 will be used for data 2.
        aggregate: Aggregate each feature by: 'mean', 'median', 'sum', 'std', 'var', 'min', 'max', or 'presence'
        **kwargs: Additional keyword arguments

    Returns:
        Plot object.
    """
    x1_dims = DataHandler.get_shape(X1)
    x2_dims = DataHandler.get_shape(X2)
    precomputed = (
        aggregate is None
        and (value_name is not None or len(x1_dims) == 1 or x1_dims[1] == 1)
        and (value_name is not None or len(x2_dims) == 1 or x2_dims[1] == 1)
    )

    if columns1 and not columns2:
        columns2 = columns1

    if aggregate == "presence":
        from biofit.stat import RowMissingnessStat

        value_name = value_name or "Presence"
        num_cols1 = x1_dims[1] if len(x1_dims) == 2 else 1
        num_cols2 = x2_dims[1] if len(x2_dims) == 2 else 1

        missingness1 = RowMissingnessStat(
            input_columns=columns1, keep_unused_columns=False, **aggregate_kwargs
        ).fit_transform(X1, output_format="pandas")
        missingness2 = RowMissingnessStat(
            input_columns=columns2, keep_unused_columns=False, **aggregate_kwargs
        ).fit_transform(X2, output_format="pandas")

        data1 = num_cols1 - missingness1
        data2 = num_cols2 - missingness2
        data1.columns = [value_name]
        data2.columns = [value_name]
    elif aggregate == "sum":
        from biofit.stat import RowSumStat

        value_name = value_name or "Sum"
        data1 = RowSumStat(
            input_columns=columns1, keep_unused_columns=False, **aggregate_kwargs
        ).fit_transform(X1, output_format="pandas")
        data2 = RowSumStat(
            input_columns=columns2, keep_unused_columns=False, **aggregate_kwargs
        ).fit_transform(X2, output_format="pandas")
        data1.columns = [value_name]
        data2.columns = [value_name]
    elif aggregate == "mean":
        from biofit.stat import RowMeanStat

        value_name = value_name or "Mean"
        data1 = RowMeanStat(
            input_columns=columns1, keep_unused_columns=False, **aggregate_kwargs
        ).fit_transform(X1, output_format="pandas")
        data2 = RowMeanStat(
            input_columns=columns2, keep_unused_columns=False, **aggregate_kwargs
        ).fit_transform(X2, output_format="pandas")
        data1.columns = [value_name]
        data2.columns = [value_name]

    value_name = value_name or "Value"

    data1 = DataHandler.to_pandas(data1)
    data2 = DataHandler.to_pandas(data2)

    if columns1:
        data1 = data1[columns1]

    if columns2:
        data2 = data2[columns2]

    if not precomputed:
        if data1.shape[1] > 1:
            data1 = data1.melt(value_name=value_name)
        else:
            data1.columns = [value_name]

        if data2.shape[1] > 1:
            data2 = data2.melt(value_name=value_name)
        else:
            data1.columns = [value_name]

    if value_name not in data1.columns:
        raise ValueError(
            f"Column '{value_name}' not found in data1. Found columns: {data1.columns}"
        )
    if value_name not in data2.columns:
        raise ValueError(
            f"Column '{value_name}' not found in data2. Found columns: {data2.columns}"
        )

    return generate_comparison_histogram(
        data1,
        data2,
        column1=value_name,
        column2=value_name,
        xlab=xlab,
        ylab=ylab,
        title=title,
        bins=bins,
        alpha=alpha,
        legend_title=legend_title,
        legend_position=legend_position,
        col_set=col_set,
        cols=cols,
        subplot_title1=subplot_title1,
        subplot_title2=subplot_title2,
        xlog=xlog,
        ylog=ylog,
        output_dir=output_dir,
        device=device,
        fingerprint=fingerprint,
        unused_columns=unused_columns,
        raise_if_missing=raise_if_missing,
    )


def plot_dimension_reduction(
    X,
    labels=None,
    group=None,
    input_columns: SelectedColumnTypes = None,
    label_column: SelectedColumnTypes = None,
    group_column: SelectedColumnTypes = None,
    method=None,
    method_kwargs={},
    title: str = Unset('"Dimension Reduction Plot"'),
    colormap: str = Unset('"nipy_spectral"'),
    n_components: int = Unset("3"),
    dimension_reducer=None,
    output_dir: str = None,
    device: str = "pdf",
    fingerprint: str = None,
    unused_columns: SelectedColumnTypes = None,
    raise_if_missing: bool = True,
    show=True,
):
    """Plot the dimension reduction plot.

    Args:
        X: Input data
        labels: Labels for the data
        group: Group for the data
        dimension_reducer: Dimension reducer object
        **kwargs: Additional keyword arguments

    Returns:
        Plot object.
    """
    requires_backends(plot_sample_metadata, ["matplotlib", "seaborn"])
    if labels is None and label_column is None:
        if "biosets" in sys.modules and isinstance(
            X, getattr(sys.modules["biosets"], "Bioset")
        ):
            from biosets import get_target

            labels = get_target(X)
    elif label_column:
        labels = DataHandler.select_columns(X, label_column)
        if "biosets" in sys.modules and isinstance(
            X, getattr(sys.modules["biosets"], "Bioset")
        ):
            from biosets import decode

            labels = decode(labels)

    if input_columns is not None:
        X = DataHandler.select_columns(X, input_columns)
        input_columns = None
    if method is None:
        from biofit.preprocessing import PCAFeatureExtractor

        method = PCAFeatureExtractor
    if isinstance(method, type):
        method = method()
    if isinstance(method, str):
        from biofit.auto.processing_auto import AutoPreprocessor

        if method_kwargs is None or not isinstance(method_kwargs, dict):
            method_kwargs = {}

        method = AutoPreprocessor.for_processor(method, **method_kwargs)
    if not method.is_fitted:
        data = method.fit_transform(X, load_from_cache_file=False)
    else:
        data = method.transform(X)

    from biofit.visualization.dimension_reduction import DimensionReductionPlotter

    return DimensionReductionPlotter().plot(
        data,
        labels,
        group,
        input_columns=input_columns,
        label_column=label_column,
        group_column=group_column,
        n_components=n_components,
        dimension_reducer=method,
        show=show,
        path=output_dir,
        device=device,
        fingerprint=fingerprint,
        unused_columns=unused_columns,
        raise_if_missing=raise_if_missing,
    )


def get_feature_importances(models, label_names=None):
    if not isinstance(models, list):
        models = [models]
    tables = []
    for fold, model in enumerate(models):
        if isinstance(model, Pipeline):
            _m = model[-1]
        else:
            _m = model
        features = _m.config.feature_names_in_
        feature_importances = _m.feature_importances_
        classes = label_names
        if classes is None:
            if hasattr(_m, "config") and hasattr(_m.config, "class_names"):
                classes = _m.config.class_names
            if classes is None:
                classes = getattr(_m, "classes_", None)

        if classes is not None and (
            (len(feature_importances) // len(classes)) == len(features)
        ):
            # this means its a one vs all classifier
            _tables = []
            for i, c in enumerate(classes):
                _tables.append(
                    pd.Series(
                        feature_importances[
                            i * len(features) : (i + 1) * len(features)
                        ],
                        index=features,
                        name=f"importances_{fold + 1}_{c}_vs_all",
                    )
                )
            tables.extend(_tables)
        else:
            tables.append(
                pd.DataFrame(
                    {
                        f"importances_{fold + 1}": feature_importances,
                    },
                    index=features,
                )
            )

    if len(tables) == 1:
        return tables[0]

    # Concatenate tables horizontally
    return pd.concat(tables, axis=1, ignore_index=False, copy=False)


def plot_feature_importance(
    feature_importances,
    models=None,
    X=None,
    y=None,
    sample_metadata=None,
    feature_metadata: dict = None,
    input_columns: SelectedColumnTypes = None,
    target_columns: SelectedColumnTypes = None,
    sample_metadata_columns: SelectedColumnTypes = None,
    sample_column: str = None,
    label_names: List[str] = None,
    plot_top: int = Unset("15"),
    feature_meta_name: str = Unset("None"),
    feature_column: str = Unset("None"),
    cols: List[str] = Unset("None"),
    colHeat: List[str] = Unset("None"),
    dat_log: str = Unset("field(default=None, init=True, repr=False)"),
    show_column_names: bool = Unset("False"),
    scale_legend_title: str = Unset('"Value"'),
    column_title: str = Unset('"Samples"'),
    row_title: str = Unset('"Features"'),
    plot_title: str = Unset('"Values"'),
    output_dir: str = None,
    device: str = "pdf",
    fingerprint: str = None,
    unused_columns: SelectedColumnTypes = None,
    raise_if_missing: bool = True,
    show=True,
):
    """Plot the feature importance of the input data.

    Args:
        X: Input data
        models: List of models
        y: Target variable
        sample_metadata: Sample metadata
        feature_metadata: Feature metadata
    Returns:
        Plot object.
    """

    from biofit.visualization.feature_importance import FeatureImportancePlotter

    if feature_importances is None and models is not None:
        feature_importances = get_feature_importances(models, label_names)
        feature_importances = feature_importances.reset_index(names=["features"])
    elif feature_importances is None:
        raise ValueError("feature_importances or models must be provided")

    FeatureImportancePlotter().plot(
        X,
        y=y,
        sample_metadata=sample_metadata,
        input_columns=input_columns,
        target_columns=target_columns,
        sample_metadata_columns=sample_metadata_columns,
        feature_importances=feature_importances,
        feature_metadata=feature_metadata,
        plot_top=plot_top,
        feature_meta_name=feature_meta_name,
        sample_column=sample_column,
        feature_column=feature_column,
        cols=cols,
        colHeat=colHeat,
        dat_log=dat_log,
        show_column_names=show_column_names,
        scale_legend_title=scale_legend_title,
        column_title=column_title,
        row_title=row_title,
        plot_title=plot_title,
        show=show,
        path=output_dir,
        device=device,
        fingerprint=fingerprint,
        unused_columns=unused_columns,
        raise_if_missing=raise_if_missing,
    )


def plot_sample_metadata(
    sample_metadata,
    outcome_data=None,
    sample_metadata_columns: Optional[SelectedColumnTypes] = None,
    outcome_column: Optional[SelectedColumnTypes] = None,
    output_dir: str = None,
    device: str = "pdf",
    fingerprint: str = None,
    unused_columns: SelectedColumnTypes = None,
    raise_if_missing: bool = True,
):
    requires_backends(plot_sample_metadata, ["rpy2"])
    from biofit.visualization.sample_metadata import SampleMetadataPlotter

    SampleMetadataPlotter().plot(
        sample_metadata,
        outcome_data,
        sample_metadata_columns=sample_metadata_columns,
        outcome_column=outcome_column,
        path=output_dir,
        device=device,
        fingerprint=fingerprint,
        unused_columns=unused_columns,
        raise_if_missing=raise_if_missing,
    )



from dataclasses import dataclass, field
from pathlib import Path

from biofit.visualization.plotting import BasePlotter, PlotterConfig


@dataclass
class SampleFilterPlotterConfig(PlotterConfig):
    processor_type: str = field(default="filtering", init=False, repr=False)
    r_source: str = field(
        default=(Path(__file__).parent / "plot_filtering.R").as_posix(),
        init=False,
        repr=False,
    )
    main_method: str = field(default="plot_filter", init=False, repr=False)


class SampleFilterPlotter(BasePlotter):
    config_class = SampleFilterPlotterConfig
    config: SampleFilterPlotterConfig



from dataclasses import dataclass, field
from typing import List, Type

from biofit.integration.biosets import get_feature
from biofit.processing import BaseProcessor, ProcessorConfig


@dataclass
class ScalerConfig(ProcessorConfig):
    processor_type: str = field(default="scaling", init=False, repr=False)
    _fit_unused_feature_types: List[Type] = field(
        default_factory=lambda: [get_feature("METADATA_FEATURE_TYPES")],
        init=False,
        repr=False,
    )
    _transform_unused_feature_types: List[Type] = field(
        default_factory=lambda: [get_feature("METADATA_FEATURE_TYPES")],
        init=False,
        repr=False,
    )


class Scaler(BaseProcessor):
    pass



from dataclasses import dataclass, field
from typing import TYPE_CHECKING, List, Optional, Type

import numpy as np
from biocore import DataHandler
from sklearn.metrics import DistanceMetric

from biofit.integration.biosets import get_feature
from biofit.processing import SelectedColumnTypes, sync_backup_config

from ..stat import Stat, StatConfig

if TYPE_CHECKING:
    pass


@dataclass
class DistanceStatConfig(StatConfig):
    """
    A base class for distance metrics.

    Inherits all attributes and methods from StatConfig.

    This class is tailored to handle distance metrics, including Minkowski,
    weighted and unweighted, seuclidean, and mahalanobis.
    """

    _fit_unused_feature_types: List[Type] = field(
        default_factory=lambda: [get_feature("METADATA_FEATURE_TYPES")],
        init=False,
        repr=False,
    )
    _transform_unused_feature_types: List[Type] = field(
        default_factory=lambda: [get_feature("METADATA_FEATURE_TYPES")],
        init=False,
        repr=False,
    )
    processor_name: str = field(default="distance", init=False, repr=False)
    metric: str = "euclidean"


@dataclass
class DistanceStatConfigForMetagenomics(DistanceStatConfig):
    """
    Calculates distance metrics specifically for metagenomics data.

    Inherits all configs from DistanceStatConfig.

    This class is tailored to handle metagenomics data, including OTU abundance,
    ASV abundance, and read counts, using the 'braycurtis' metric by default.
    """

    _fit_input_feature_types: List[Type] = field(
        default_factory=lambda: [(get_feature("Abundance"), get_feature("ReadCount"))],
        init=False,
        repr=False,
    )
    _transform_input_feature_types: List[Type] = field(
        default_factory=lambda: [(get_feature("Abundance"), get_feature("ReadCount"))],
        init=False,
        repr=False,
    )
    dataset_name: str = field(default="metagenomics", init=False, repr=False)
    metric: str = "braycurtis"


@dataclass
class DistanceStatConfigForOTU(DistanceStatConfig):
    """
    A subclass for calculating distance metrics on OTU abundance data.

    Inherits all configs from DistanceStatConfig, but is specifically
    tailored for OTU abundance data, defaulting to the 'braycurtis' metric.
    """

    _fit_input_feature_types: List[Type] = field(
        default_factory=lambda: [get_feature("Abundance")], init=False, repr=False
    )
    _transform_input_feature_types: List[Type] = field(
        default_factory=lambda: [get_feature("Abundance")], init=False, repr=False
    )
    dataset_name: str = field(default="otu", init=False, repr=False)
    metric: str = "braycurtis"


@dataclass
class DistanceStatConfigForASV(DistanceStatConfig):
    """
    A subclass for calculating distance metrics on ASV abundance data.

    Inherits all attributes and methods from DistanceStat, but is specifically
    tailored for ASV abundance data, defaulting to the 'braycurtis' metric.
    """

    _fit_input_feature_types: List[Type] = field(
        default_factory=lambda: [get_feature("Abundance")], init=False, repr=False
    )
    _transform_input_feature_types: List[Type] = field(
        default_factory=lambda: [get_feature("Abundance")], init=False, repr=False
    )
    dataset_name: str = field(default="asv", init=False, repr=False)
    metric: str = "braycurtis"


class DistanceStatConfigForReadCount(DistanceStatConfig):
    """
    A subclass for calculating distance metrics on read count data.

    Inherits all attributes and methods from DistanceStat, but is specifically
    tailored for read count data, defaulting to the 'braycurtis' metric.
    """

    _fit_input_feature_types: List[Type] = field(
        default_factory=lambda: [get_feature("ReadCount")], init=False, repr=False
    )
    _transform_input_feature_types: List[Type] = field(
        default_factory=lambda: [get_feature("ReadCount")], init=False, repr=False
    )
    dataset_name: str = field(default="read_count", init=False, repr=False)
    metric: str = "braycurtis"


class DistanceStatConfigForSNP(DistanceStatConfig):
    """
    A subclass for calculating distance metrics on read count data.

    Inherits all attributes and methods from DistanceStat, but is specifically
    tailored for read count data, defaulting to the 'braycurtis' metric.
    """

    _fit_input_feature_types: List[Type] = field(
        default_factory=lambda: [get_feature("GenomicVariant")], init=False, repr=False
    )
    _transform_input_feature_types: List[Type] = field(
        default_factory=lambda: [get_feature("GenomicVariant")], init=False, repr=False
    )
    dataset_name: str = field(default="snp", init=False, repr=False)

    metric: str = "jaccard"


class DistanceStat(Stat):
    """
    A base class for calculating distance metrics on genomic data.

    Attributes:
        metric (str): The name of the distance metric to use. Defaults to 'euclidean'.
        p (float): The p-norm to apply for Minkowski, weighted and unweighted. Default is 2.
        w (Union[None, np.ndarray], optional): The weight vector for weighted Minkowski. Default is None.
        V (Union[None, np.ndarray], optional): The variance vector for seuclidean. Default is None.
        VI (Union[None, np.ndarray], optional): The inverse of the covariance matrix for mahalanobis. Default is None.

    Methods:
        fit_sklearn(X: Union[np.ndarray, pd.DataFrame, "pl.DataFrame"]): Validates input data and calculates the pairwise distances between samples in X.
    """

    _config_class = DistanceStatConfig
    config: DistanceStatConfig
    output_dtype = "float64"

    def __init__(
        self,
        config: Optional[DistanceStatConfig] = None,
        metric: Optional[str] = "euclidean",
        **kwargs,
    ):
        super().__init__(config=config, metric=metric, **kwargs)
        self.pdist = DistanceMetric.get_metric(self.config.metric).pairwise

    @sync_backup_config
    def set_params(self, **kwargs):
        self.config = self.config.replace_defaults(**kwargs)
        self.pdist = DistanceMetric.get_metric(self.config.metric).pairwise
        return self

    def fit(
        self,
        X,
        input_columns: SelectedColumnTypes = None,
        raise_if_missing: bool = None,
        cache_output: bool = None,
        load_from_cache_file: bool = None,
        batched: bool = True,
        batch_size: int = 1000,
        batch_format: str = None,
        num_proc: int = None,
        map_kwargs: dict = {"fn_kwargs": {}},
        cache_dir: str = None,
        cache_file_name: str = None,
        fingerprint: str = None,
    ) -> "DistanceStat":
        self.config._input_columns = self._set_input_columns_and_arity(input_columns)
        return self._process_fit(
            X,
            raise_if_missing=raise_if_missing,
            cache_output=cache_output,
            load_from_cache_file=load_from_cache_file,
            batched=batched,
            batch_size=batch_size,
            batch_format=batch_format,
            num_proc=num_proc,
            map_kwargs=map_kwargs,
            cache_dir=cache_dir,
            cache_file_name=cache_file_name,
            fingerprint=fingerprint,
        )

    def transform(
        self,
        X,
        input_columns: SelectedColumnTypes = None,
        keep_unused_columns: bool = None,
        raise_if_missing: bool = None,
        cache_output: bool = None,
        cache_dir: str = None,
        cache_file_name: str = None,
        load_from_cache_file: bool = None,
        batched: bool = None,
        batch_size: int = None,
        batch_format: str = None,
        output_format: str = None,
        map_kwargs: dict = None,
        num_proc: int = None,
        fingerprint: str = None,
    ):
        self.config._n_features_out = DataHandler.get_shape(X)[0]
        self._input_columns = self._set_input_columns_and_arity(input_columns)
        return self._process_transform(
            X,
            keep_unused_columns=keep_unused_columns,
            raise_if_missing=raise_if_missing,
            cache_output=cache_output,
            cache_dir=cache_dir,
            cache_file_name=cache_file_name,
            load_from_cache_file=load_from_cache_file,
            batched=batched,
            batch_size=batch_size,
            batch_format=batch_format,
            output_format=output_format,
            map_kwargs=map_kwargs,
            num_proc=num_proc,
            fingerprint=fingerprint,
        )

    def fit_transform(
        self,
        X,
        input_columns: SelectedColumnTypes = None,
        keep_unused_columns: bool = True,
        raise_if_missing: bool = False,
        cache_output: bool = None,
        load_from_cache_file: bool = None,
        batched: bool = True,
        batch_size: int = 1000,
        batch_format: str = None,
        output_format: str = None,
        map_kwargs: dict = {"fn_kwargs": {}},
        num_proc: int = None,
        cache_dir: str = None,
        cache_file_name: str = None,
        fingerprint: str = None,
    ):
        return self.fit(
            X,
            input_columns=input_columns,
            raise_if_missing=raise_if_missing,
            cache_output=cache_output,
            load_from_cache_file=load_from_cache_file,
            batched=batched,
            batch_size=batch_size,
            batch_format=batch_format,
            num_proc=num_proc,
            map_kwargs=map_kwargs,
            cache_dir=cache_dir,
            cache_file_name=cache_file_name,
            fingerprint=fingerprint,
        ).transform(
            X,
            input_columns=input_columns,
            keep_unused_columns=keep_unused_columns,
            raise_if_missing=raise_if_missing,
            cache_output=cache_output,
            cache_dir=cache_dir,
            cache_file_name=cache_file_name,
            load_from_cache_file=load_from_cache_file,
            batched=batched,
            batch_size=batch_size,
            batch_format=batch_format,
            output_format=output_format,
            map_kwargs=map_kwargs,
            num_proc=num_proc,
            fingerprint=fingerprint,
        )

    def _transform_numpy(self, X: np.ndarray):
        return self.pdist(X, X)



from setuptools import find_packages, setup

REQUIRED_PKGS = [
    # the library that biofit is built upon
    "biocore>=1.1.1",
    # For file locking
    "filelock",
    # We use numpy>=1.17 to have np.random.Generator (Dataset shuffling)
    "numpy>=1.17",
    # Backend and serialization.
    # Minimum 8.0.0 to be able to use .to_reader()
    "pyarrow>=8.0.0",
    # As long as we allow pyarrow < 14.0.1, to fix vulnerability CVE-2023-47248
    "pyarrow-hotfix",
    # For smart caching dataset processing
    "dill>=0.3.0,<0.3.8",  # tmp pin until dill has official support for determinism see https://github.com/uqfoundation/dill/issues/19
    # For performance gains with apache arrow
    "pandas",
    # for downloading datasets over HTTPS
    "requests>=2.19.0",
    # progress bars in download and scripts
    "tqdm>=4.62.1",
    # for fast hashing
    "xxhash",
    # for better multiprocessing
    "multiprocess",
    # to save datasets locally or on any filesystem
    # minimum 2023.1.0 to support protocol=kwargs in fsspec's `open`, `get_fs_token_paths`, etc.: see https://github.com/fsspec/filesystem_spec/pull/1143
    "fsspec[http]>=2023.1.0,<=2023.10.0",
    # Utilities from PyPA to e.g., compare versions
    "packaging",
    # To parse YAML metadata from dataset cards
    "pyyaml>=5.1",
    # for processing and transforming datasets
    "scikit-learn",
]

QUALITY_REQUIRE = ["ruff>=0.1.5"]

DOCS_REQUIRE = [
    # Might need to add doc-builder and some specific deps in the future
    "s3fs",
]

VISUALIZATION_REQUIRE = [
    "matplotlib",
    "seaborn",
    "psutil",  # for the start time of the biofit run
]


ML_REQUIRE = [
    "polars>=0.20.5",
    "timezones>=0.10.2",
    "optuna",
    "lightgbm",
    # "xgboost",
    # "catboost",
    "imbalanced-learn",
]

TESTS_REQUIRE = ["pytest", "pytest-timeout", "pytest-xdist"]


EXTRAS_REQUIRE = {
    "polars": ["polars>=0.20.5", "timezones>=0.10.2"],
    "rpy2": ["rpy2>=3.5.15", "rpy2-arrow>=0.0.8"],
    "ml": ML_REQUIRE,
    "apache-beam": ["apache-beam>=2.26.0,<2.44.0"],
    "vcf": ["cyvcf2>=0.30.0", "sgkit>=0.0.1"],
    "tensorflow": [
        "tensorflow>=2.2.0,!=2.6.0,!=2.6.1; sys_platform != 'darwin' or platform_machine != 'arm64'",
        "tensorflow-macos; sys_platform == 'darwin' and platform_machine == 'arm64'",
    ],
    "tensorflow_gpu": ["tensorflow-gpu>=2.2.0,!=2.6.0,!=2.6.1"],
    "torch": ["torch"],
    "jax": ["jax>=0.3.14", "jaxlib>=0.3.14"],
    "s3": ["s3fs"],
    "viz": VISUALIZATION_REQUIRE,
    "test": QUALITY_REQUIRE
    + TESTS_REQUIRE
    + DOCS_REQUIRE
    + VISUALIZATION_REQUIRE
    + ML_REQUIRE,
    "all": VISUALIZATION_REQUIRE + ML_REQUIRE + QUALITY_REQUIRE + DOCS_REQUIRE,
    "quality": QUALITY_REQUIRE,
    "docs": DOCS_REQUIRE,
}

setup(
    name="biofit",
    version="0.0.1",  # expected format is one of x.y.z.dev0, or x.y.z.rc1 or x.y.z (no to dashes, yes to dots)
    description="BioFit: Bioinformatics Machine Learning Framework",
    long_description=open("README.md", encoding="utf-8").read(),
    long_description_content_type="text/markdown",
    author="Patrick Smyth",
    author_email="psmyth1994@gmail.com",
    url="https://github.com/psmyth94/biofit",
    download_url="https://github.com/psmyth94/biofit/tags",
    license="MIT",
    package_dir={"": "src"},
    packages=find_packages("src"),
    include_package_data=True,
    python_requires=">=3.8.0,<3.12.0",
    install_requires=REQUIRED_PKGS,
    extras_require=EXTRAS_REQUIRE,
    entry_points={
        "console_scripts": [
            "biofit = biofit.cli.main:main",
        ],
    },
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Developers",
        "Intended Audience :: Education",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Topic :: Scientific/Engineering :: Artificial Intelligence",
        "Topic :: Scientific/Engineering :: Bio-Informatics",
    ],
    keywords="omics machine learning bioinformatics metrics",
    zip_safe=False,  # Required for mypy to find the py.typed file
)



import textwrap
from dataclasses import dataclass, field
from typing import List, Optional, Type

from biocore import DataHandler
from biocore.utils.py_util import is_bioset

import biofit.config as config
from biofit.integration.biosets import get_feature
from biofit.integration.R import RCaller
from biofit.processing import SelectedColumnTypes, sync_backup_config
from biofit.utils.types import Unset
from biofit.visualization.plotting import BasePlotter, PlotterConfig


@dataclass
class BarPlotConfig(PlotterConfig):
    _fit_input_feature_types: List[Type] = field(
        default_factory=lambda: [None, get_feature("TARGET_FEATURE_TYPES"), None],
        init=False,
        repr=False,
    )
    _transform_input_feature_types: List[Type] = field(
        default_factory=lambda: [None, get_feature("TARGET_FEATURE_TYPES"), None],
        init=False,
        repr=False,
    )

    _fit_unused_feature_types: List[Type] = field(
        default_factory=lambda: [get_feature("METADATA_FEATURE_TYPES"), None, None],
        init=False,
        repr=False,
    )
    _transform_unused_feature_types: List[Type] = field(
        default_factory=lambda: [get_feature("METADATA_FEATURE_TYPES"), None, None],
        init=False,
        repr=False,
    )

    processor_type: str = field(default="scaling", init=False, repr=False)
    r_source: str = field(
        default=(config.R_SCRIPTS / "plotting_utils.R").as_posix(),
        init=False,
        repr=False,
    )
    main_method: str = field(default="generate_barplot", init=False, repr=False)

    label_name: str = None
    value_name: Optional[str] = None
    groupby: Optional[str] = None
    xlab: Optional[str] = None
    ylab: Optional[str] = None
    title: str = "Bar Plot"
    col_set: str = "Set1"
    col_outline: str = "grey30"
    col_labels: str = "black"
    cols: Optional[str] = None
    prop: bool = False
    add_count_lab: bool = True
    vars_as_entered: bool = False
    legend_position: str = "top"
    font_size: float = 3.25


class BarPlotter(BasePlotter):
    _config_class = BarPlotConfig
    config: BarPlotConfig

    def __init__(
        self,
        xlab: Optional[str] = None,
        ylab: Optional[str] = None,
        title: str = Unset('"Bar Plot"'),
        col_set: str = Unset('"Set1"'),
        col_labels: str = Unset('"black"'),
        col_outline: str = Unset('"grey30"'),
        cols: Optional[List[str]] = Unset("None"),
        prop: bool = Unset("False"),
        add_count_lab: bool = Unset("True"),
        vars_as_entered: bool = Unset("False"),
        legend_position: str = Unset('"top"'),
        font_size: float = Unset("3.25"),
        path=None,
        config: Optional[BarPlotConfig] = None,
    ):
        super().__init__(
            config=config,
            xlab=xlab,
            ylab=ylab,
            title=title,
            col_set=col_set,
            col_labels=col_labels,
            cols=cols,
            prop=prop,
            add_count_lab=add_count_lab,
            vars_as_entered=vars_as_entered,
            legend_position=legend_position,
            font_size=font_size,
            path=path,
        )

        self = self.__post_init__()

    def __post_init__(self):
        if self.config.r_source and self.config.main_method:
            self.r_caller = RCaller.from_script(self.config.r_source)
            self.r_caller.verify_r_dependencies()
            enter_code = textwrap.dedent(
                """
                suppressPackageStartupMessages(require(ggplot2))
                """
            )
            exit_code = textwrap.dedent(
                """
                ggplot2::ggsave(path, results)
                """
            )
            self.plotter = self.r_caller.get_method(
                self.config.main_method, enter_code, exit_code
            )

        return self

    @sync_backup_config
    def set_params(self, **kwargs):
        self.config = self.config.replace_defaults(**kwargs)
        self = self.__post_init__()
        return self

    def plot(
        self,
        x,
        y,
        group,
        label_name: str = None,
        value_name: Optional[str] = None,
        groupby: Optional[str] = None,
        xlab: Optional[str] = Unset("None"),
        ylab: Optional[str] = Unset("None"),
        title: str = Unset('"Bar Plot"'),
        col_set: str = Unset('"Set1"'),
        col_labels: str = Unset('"black"'),
        col_outline: str = Unset('"grey30"'),
        cols: Optional[List[str]] = Unset("None"),
        prop: bool = Unset("False"),
        add_count_lab: bool = Unset("True"),
        vars_as_entered: bool = Unset("False"),
        legend_position: str = Unset('"top"'),
        font_size: float = Unset("3.25"),
        path: str = None,
        device: str = "pdf",
        fingerprint: str = None,
        unused_columns: SelectedColumnTypes = None,
        raise_if_missing: bool = True,
        show: bool = True,
    ):
        self.config._input_columns = self._set_input_columns_and_arity(
            value_name, label_name, groupby
        )
        return self._plot(
            x,
            y,
            group,
            xlab=xlab,
            ylab=ylab,
            title=title,
            col_set=col_set,
            col_labels=col_labels,
            col_outline=col_outline,
            cols=cols,
            prop=prop,
            add_count_lab=add_count_lab,
            vars_as_entered=vars_as_entered,
            legend_position=legend_position,
            font_size=font_size,
            path=path,
            device=device,
            fingerprint=fingerprint,
            unused_columns=unused_columns,
            raise_if_missing=raise_if_missing,
            show=show,
        )

    def plot_dataset(self, x, y, group):
        if is_bioset(x):
            from biosets import decode

            x = decode(x)
        if is_bioset(group):
            group = decode(group)

        return self.plot_arrow(
            DataHandler.to_arrow(x),
            DataHandler.to_arrow(y) if y is not None else None,
            DataHandler.to_arrow(group) if group is not None else None,
        )

    def plot_arrow(self, x, y=None, group=None):
        kwargs = self.config.get_params()
        self.plotter(x, y, group, **kwargs)



# ruff: noqa
from .configuration_auto import (
    AutoConfig,
    AutoPreprocessorConfig,
)
from .modeling_auto import (
    AutoModel,
    AutoModelForClassification,
)
from .processing_auto import AutoProcessor, AutoPreprocessor
from .plotting_auto import AutoPlotter, PlotterPipeline



from dataclasses import dataclass, field
from typing import List, Type

import numpy as np
from biocore import DataHandler

from biofit.integration.biosets import get_feature
from biofit.processing import SelectedColumnTypes, sync_backup_config
from biofit.utils import logging

from ..encoding import Encoder, EncoderConfig

logger = logging.get_logger(__name__)


@dataclass
class LabelEncoderConfig(EncoderConfig):
    """Configuration class for label encoding."""

    _fit_process_desc: str = field(
        default="Determining unique labels", init=False, repr=False
    )
    _fit_input_feature_types: List[Type] = field(
        default_factory=lambda: [get_feature("ClassLabel")], init=False, repr=False
    )
    _transform_input_feature_types: List[Type] = field(
        default_factory=lambda: [get_feature("ClassLabel")], init=False, repr=False
    )

    processor_name: str = field(default="label_encoding", init=False, repr=False)
    _transform_process_desc: str = field(default=None, init=False, repr=False)

    names: list = field(default_factory=list, init=False, repr=False)

    # auto attributes
    label_mapping: dict = field(default_factory=dict, init=False, repr=False)

    def __post_init__(self):
        if self.names:
            self._transform_process_desc = f"Encoding labels with {self.names}"
            self.label_mapping = {label: i for i, label in enumerate(self.names)}


class LabelEncoder(Encoder):
    config_class = LabelEncoderConfig
    config: LabelEncoderConfig
    output_feature_type = get_feature("ClassLabel")

    @sync_backup_config
    def set_params(self, **kwargs):
        self.config = self.config.replace_defaults(**kwargs)
        self.config.__post_init__()
        return self

    def fit(
        self,
        X,
        input_columns: SelectedColumnTypes = None,
        raise_if_missing: bool = None,
        cache_output: bool = None,
        load_from_cache_file: bool = None,
        batched: bool = True,
        batch_size: int = 1000,
        batch_format: str = None,
        num_proc: int = None,
        map_kwargs: dict = {"fn_kwargs": {}},
        cache_dir: str = None,
        cache_file_name: str = None,
        fingerprint: str = None,
    ) -> "LabelEncoder":
        self.config._input_columns = self._set_input_columns_and_arity(input_columns)
        return self._process_fit(
            X,
            input_columns=input_columns,
            raise_if_missing=raise_if_missing,
            cache_output=cache_output,
            load_from_cache_file=load_from_cache_file,
            batched=batched,
            batch_size=batch_size,
            batch_format=batch_format,
            num_proc=num_proc,
            map_kwargs=map_kwargs,
            cache_dir=cache_dir,
            cache_file_name=cache_file_name,
            fingerprint=fingerprint,
        )

    def transform(
        self,
        X,
        input_columns: SelectedColumnTypes = None,
        keep_unused_columns: bool = None,
        raise_if_missing: bool = None,
        cache_output: bool = None,
        cache_dir: str = None,
        cache_file_name: str = None,
        load_from_cache_file: bool = None,
        batched: bool = None,
        batch_size: int = None,
        batch_format: str = None,
        output_format: str = None,
        map_kwargs: dict = None,
        num_proc: int = None,
        fingerprint: str = None,
    ):
        self._input_columns = self._set_input_columns_and_arity(input_columns)
        return self._process_transform(
            X,
            keep_unused_columns=keep_unused_columns,
            raise_if_missing=raise_if_missing,
            cache_output=cache_output,
            cache_dir=cache_dir,
            cache_file_name=cache_file_name,
            load_from_cache_file=load_from_cache_file,
            batched=batched,
            batch_size=batch_size,
            batch_format=batch_format,
            output_format=output_format,
            map_kwargs=map_kwargs,
            num_proc=num_proc,
            fingerprint=fingerprint,
        )

    def fit_transform(
        self,
        X,
        *args,
        input_columns: SelectedColumnTypes = None,
        keep_unused_columns: bool = True,
        raise_if_missing: bool = False,
        cache_output: bool = None,
        load_from_cache_file: bool = None,
        batched: bool = True,
        batch_size: int = 1000,
        batch_format: str = None,
        output_format: str = None,
        map_kwargs: dict = {"fn_kwargs": {}},
        num_proc: int = None,
        cache_dir: str = None,
        cache_file_name: str = None,
        fingerprint: str = None,
    ):
        return self.fit(
            X,
            input_columns=input_columns,
            raise_if_missing=raise_if_missing,
            cache_output=cache_output,
            load_from_cache_file=load_from_cache_file,
            batched=batched,
            batch_size=batch_size,
            batch_format=batch_format,
            num_proc=num_proc,
            map_kwargs=map_kwargs,
            cache_dir=cache_dir,
            cache_file_name=cache_file_name,
            fingerprint=fingerprint,
        ).transform(
            X,
            input_columns=input_columns,
            keep_unused_columns=keep_unused_columns,
            raise_if_missing=raise_if_missing,
            cache_output=cache_output,
            cache_dir=cache_dir,
            cache_file_name=cache_file_name,
            load_from_cache_file=load_from_cache_file,
            batched=batched,
            batch_size=batch_size,
            batch_format=batch_format,
            output_format=output_format,
            map_kwargs=map_kwargs,
            num_proc=num_proc,
            fingerprint=fingerprint,
        )

    def _get_features_out(self, X, **kwargs):
        features = super()._get_features_out(X, **kwargs)
        if features:
            input_names = self.config.feature_names_in_ or self.config.feature_idx_in_
            for name in input_names:
                label_feature = features.get(name)
                if isinstance(label_feature, get_feature("ClassLabel")):
                    label_names = label_feature.names
                    if self.config.names and isinstance(self.config.names[0], str):
                        self.config.label_mapping = {
                            label_names.index(label): i
                            for i, label in enumerate(self.config.names)
                        }

                label_feature = get_feature("ClassLabel")(names=self.config.names)
                features[name] = label_feature

        return features

    def _check_data(self, X):
        X_dims = DataHandler.get_shape(X)
        if len(X_dims) > 1:
            if X_dims[1] == 1:
                out = DataHandler.select_column(X, 0)
            else:
                raise ValueError(
                    f"Expected input to have 1 column, got {X_dims[1]} columns"
                )
        else:
            out = X
        return out

    def _process_fit_input(self, input, **kwargs):
        if self.config.names:
            kwargs["fn_kwargs"]["fn"] = None
        return super()._process_fit_input(input, **kwargs)

    def _fit_array(self, X):
        out = self._check_data(X)
        self.config.names = DataHandler.to_format(DataHandler.unique(out), "list")
        self.config._transform_process_desc = (
            f"Encoding labels with: {self.config.label_mapping}"
        )
        return self

    def _partial_fit_array(self, X):
        out = self._check_data(X)
        labs = DataHandler.to_format(DataHandler.unique(out), "list")
        self.config.names = np.unique(self.config.names + labs).tolist()
        return self

    def _pool_fit(self, fitted_processors):
        names = []
        for processor in fitted_processors:
            if isinstance(processor, LabelEncoder):
                names += processor.config.names
        self.config.names = np.unique(names).tolist()
        self.config.label_mapping = {
            label: i for i, label in enumerate(self.config.names)
        }
        self.config._transform_process_desc = (
            f"Encoding labels with: {self.config.label_mapping}"
        )
        return self

    def _transform_array(self, X):
        labs = self._check_data(X)
        labs = DataHandler.to_format(labs, "list")
        return [self.config.label_mapping.get(lab, -1) for lab in labs]



import gc
import inspect
import io
import os
import re
import shutil
import subprocess
import sys
from glob import glob
from pathlib import Path
from typing import TYPE_CHECKING, List, Optional, TextIO

import numpy as np
from biocore.utils.import_util import (
    is_rpy2_arrow_available,
    is_rpy2_available,
    requires_backends,
)

import biofit.config
from biofit import config
from biofit.utils import logging

if TYPE_CHECKING:
    from rpy2.robjects.conversion import Converter


class PackageNotInstalledError(ImportError):
    """Error occuring because the R package to import is not installed."""

    pass


logger = logging.get_logger(__name__)

_is_auto_install = None

R_PLOTTING_DEPENDENCIES = {
    "cran": [
        "ggplot2",
        "arrow",
        "circlize",
        "RColorBrewer",
        "scales",
        "forcats",
        "patchwork",
        "reshape2",
        "dplyr",
        "tools",
    ],
    "bioconductor": ["ComplexHeatmap"],
}
R_PREPROCESSING_DEPENDENCIES = {
    "cran": [],
    "bioconductor": ["edgeR"],
}


def is_auto_install():
    global _is_auto_install
    if _is_auto_install is None:
        _is_auto_install = not biofit.config.BIOFIT_SKIP_R_DEPENDENCIES
    return _is_auto_install


def enable_auto_install():
    global _is_auto_install
    _is_auto_install = True


def disable_auto_install():
    global _is_auto_install
    _is_auto_install = False


class ROutputCapture:
    def __init__(self, stdout: TextIO = io.StringIO(), stderr: TextIO = io.StringIO()):
        requires_backends("ROutputCapture", "rpy2")
        from rpy2.rinterface_lib import callbacks

        # Create StringIO buffers to capture output
        self.stdout = stdout
        self.stderr = stderr
        # Save original R console write functions
        self._original_consolewrite_print = callbacks.consolewrite_print
        self._original_consolewrite_warnerror = callbacks.consolewrite_warnerror

    def __enter__(self):
        from rpy2.rinterface_lib import callbacks

        # Define new console write functions
        def custom_consolewrite_print(output):
            self.stdout.write(output)

        def custom_consolewrite_warnerror(output):
            self.stderr.write(output)

        # Replace R console write functions with custom functions
        callbacks.consolewrite_print = custom_consolewrite_print
        callbacks.consolewrite_warnerror = custom_consolewrite_warnerror

        return self

    def __exit__(self, exc_type, exc_value, traceback):
        from rpy2.rinterface_lib import callbacks

        # Restore original R console write functions
        callbacks.consolewrite_print = self._original_consolewrite_print
        callbacks.consolewrite_warnerror = self._original_consolewrite_warnerror

        # Reset the position of StringIO buffers to the beginning
        self.stdout.seek(0)
        self.stderr.seek(0)

    def get_stdout(self):
        return self.stdout.getvalue()

    def get_stderr(self):
        return self.stderr.getvalue()


def get_linux_distro_codename():
    """
    Retrieves the codename of the Linux distribution by parsing /etc/os-release.

    Returns:
        str: The codename of the distribution (e.g., 'focal', 'bullseye', 'centos7').
             Returns 'Unknown' if the codename cannot be determined.
    """
    os_release_path = "/etc/os-release"

    if not os.path.isfile(os_release_path):
        return "Unknown"

    os_info = {}

    try:
        with open(os_release_path, "r") as file:
            for line in file:
                # Remove any leading/trailing whitespace and skip empty lines
                line = line.strip()
                if not line or "=" not in line:
                    continue

                key, value = line.split("=", 1)
                # Remove surrounding quotes if present
                value = value.strip('"').strip("'")
                os_info[key] = value
    except Exception as e:
        print(f"Error reading {os_release_path}: {e}")
        return None

    # Handle specific distributions
    distro_id = os_info.get("ID", "").lower()

    if distro_id in ["ubuntu", "debian"]:
        # For Ubuntu and Debian, extract codename from VERSION or VERSION_ID
        if "VERSION_CODENAME" in os_info:
            return os_info["VERSION_CODENAME"].lower()
        elif "VERSION" in os_info:
            # Example VERSION: "20.04.5 LTS (Focal Fossa)"
            match = re.search(r"\(([^)]+)\)", os_info["VERSION"])
            if match:
                return match.group(1).lower()
        elif "VERSION_ID" in os_info:
            return os_info["VERSION_ID"].lower()

    elif distro_id in ["centos", "rhel", "fedora"]:
        # For CentOS, RHEL, Fedora, use VERSION_ID or similar
        if "VERSION_ID" in os_info:
            return f"{distro_id}{os_info['VERSION_ID'].lower()}"
        elif "VERSION" in os_info:
            # Example VERSION: "7 (Core)"
            match = re.search(r"(\d+)", os_info["VERSION"])
            if match:
                return f"{distro_id}{match.group(1)}"

    # Attempt to get VERSION_CODENAME directly
    if "VERSION_CODENAME" in os_info:
        return os_info["VERSION_CODENAME"].lower()

    # Fallback: Try to extract codename from PRETTY_NAME or other fields
    for key in ["PRETTY_NAME", "NAME", "DESCRIPTION"]:
        if key in os_info:
            match = re.search(r"\b([a-z]+)\b", os_info[key], re.IGNORECASE)
            if match:
                return match.group(1).lower()

    return None


def get_cran_info():
    """
    Detects the R version using rpy2 and constructs the appropriate CRAN URL based on the operating system
    and, for Linux, the distribution codename.

    Returns:
        str: The constructed CRAN URL, or a default 'latest' URL if R version cannot be detected.
    """
    cran_url = None

    from rpy2.rinterface_lib.embedded import RRuntimeError

    try:
        if sys.platform in ["win32", "darwin"]:
            # For Windows, include R version in the CRAN URL
            cran_url = "https://packagemanager.posit.co/cran/latest"
        elif sys.platform.startswith("linux"):
            # For Linux and macOS
            dist = get_linux_distro_codename()
            if dist:
                cran_url = (
                    f"https://packagemanager.posit.co/cran/__linux__/{dist}/latest"
                )
            else:
                logger.warning(
                    "Linux distribution codename could not be determined. Using default CRAN URL."
                )
                return None
        else:
            logger.warning("Unknown operating system. Using default CRAN URL.")
            return None

    except RRuntimeError as e:
        logger.error(f"Error detecting R version: {e}. Using default CRAN URL.")

    return cran_url


def get_bioconductor_info():
    bioc_mirror = "https://packagemanager.posit.co/bioconductor/latest"
    bioconductor_config_file = f"{bioc_mirror}/config.yaml"
    logger.info(f"Using Bioconductor mirror: {bioc_mirror}")
    logger.info(f"Using Bioconductor config file: {bioconductor_config_file}")

    return bioconductor_config_file, bioc_mirror


class RCaller:
    r_code = None
    r_source = None
    _r_context = {}
    _global_vars = {}

    def _get_converter(self, obj, use_arrow=True) -> "Converter":
        """Check if rpy2 is available and retrieves the right converter.

        Args:
            obj (obj): The object (or name) calling this method.

        Raises:
            RuntimeError: If rpy2 is not available.

        Returns:
            Converter: The converter to convert R objects to Python objects.
        """
        if is_rpy2_arrow_available() and use_arrow:
            from rpy2.robjects import default_converter, numpy2ri, pandas2ri
            from rpy2_arrow.arrow import converter

            _converter = (
                default_converter + numpy2ri.converter + pandas2ri.converter + converter
            )
        elif is_rpy2_available():
            from rpy2.robjects import conversion, numpy2ri, pandas2ri

            _converter = (
                (
                    conversion.get_conversion()
                    if getattr(conversion, "get_conversion", None)
                    else conversion.converter
                )
                + numpy2ri.converter
                + pandas2ri.converter
            )
        else:
            # suggest installing rpy2_arrow if rpy2 is not available
            requires_backends(obj, "rpy2_arrow")
            _converter = None

        return _converter

    @staticmethod
    def get_r_home():
        """
        Retrieve the R_HOME directory across multiple systems, prioritizing local R installations,
        particularly those within an active Conda environment.

        Returns:
            str or None: The path to R_HOME if found, otherwise None.
        """
        # Step 1: Check R_HOME environment variable
        r_home_env = os.environ.get("R_HOME")
        if r_home_env and os.path.exists(r_home_env):
            return r_home_env

        # Step 2: Check for R in active Conda environment
        conda_prefix = os.environ.get("CONDA_PREFIX")
        if conda_prefix:
            if sys.platform == "win32":
                r_executable = os.path.join(conda_prefix, "Scripts", "R.exe")
            else:
                r_executable = os.path.join(conda_prefix, "bin", "R")

            if os.path.exists(r_executable):
                try:
                    r_home = subprocess.check_output(
                        [r_executable, "RHOME"], universal_newlines=True
                    ).strip()
                    if os.path.exists(r_home):
                        return r_home
                except subprocess.CalledProcessError:
                    pass  # R executable found but failed to get R_HOME

        # Step 3: Check system PATH for R executable
        r_executable = shutil.which("R")
        if r_executable:
            try:
                r_home = subprocess.check_output(
                    [r_executable, "RHOME"], universal_newlines=True
                ).strip()
                if os.path.exists(r_home):
                    return r_home
            except subprocess.CalledProcessError:
                pass  # R executable found but failed to get R_HOME

        # Step 4: Check common installation paths (additional step for thoroughness)
        potential_paths = []
        if sys.platform == "win32":
            # Windows common installation paths
            potential_paths.extend(glob("C:/Program Files/R/R-*/"))

            # Attempt to read R_HOME from Windows Registry
            try:
                import winreg

                reg_path = r"SOFTWARE\R-core\R"
                for root in [winreg.HKEY_LOCAL_MACHINE, winreg.HKEY_CURRENT_USER]:
                    try:
                        key = winreg.OpenKey(root, reg_path)
                        r_home, _ = winreg.QueryValueEx(key, "InstallPath")
                        if os.path.exists(r_home):
                            return r_home
                    except FileNotFoundError:
                        continue
            except ImportError:
                pass  # winreg not available
        else:
            # macOS and Linux common installation paths
            potential_paths.extend(
                [
                    "/usr/local/lib/R",
                    "/usr/lib/R",
                    "/Library/Frameworks/R.framework/Resources",  # macOS
                ]
            )

        for path in potential_paths:
            if os.path.exists(path):
                return path

        # R_HOME could not be determined
        return None

    @staticmethod
    def verify_r_dependencies(
        cran_dependencies: List[str] = R_PLOTTING_DEPENDENCIES["cran"]
        + R_PREPROCESSING_DEPENDENCIES["cran"],
        bioconductor_dependencies: List[str] = R_PREPROCESSING_DEPENDENCIES[
            "bioconductor"
        ]
        + R_PLOTTING_DEPENDENCIES["bioconductor"],
        install_missing: Optional[bool] = None,
        **kwargs,
    ):
        if install_missing is None:
            install_missing = is_auto_install()

        r_home = None
        try:
            import rpy2.situation

            r_home = rpy2.situation.get_r_home()
        except Exception:
            pass
        r_home = r_home or RCaller.get_r_home()
        if r_home is None:
            raise RuntimeError("R_HOME could not be determined.")

        import rpy2.robjects.packages as rpackages
        from rpy2.robjects import ListVector
        from rpy2.robjects.vectors import StrVector

        utils = rpackages.importr("utils")
        base = rpackages.importr("base")
        bioc_missing, cran_missing = [], []
        cran_url = get_cran_info()
        if cran_url is not None and "repos" not in kwargs:
            kwargs["repos"] = ListVector([("CRAN", cran_url)])
        if cran_dependencies:
            names_to_install = [
                x for x in cran_dependencies if not rpackages.isinstalled(x)
            ]
            if len(names_to_install) > 0 and install_missing:
                logger.info(f"Using R_HOME: {r_home}")
                logger.info(f"Installing missing CRAN dependencies: {names_to_install}")
                if "BiocManager" in names_to_install:
                    names_to_install.remove("BiocManager")
                    utils.chooseCRANmirror(ind=1)
                    utils.install_packages("BiocManager")

                if cran_url is not None:
                    logger.info("Using CRAN mirror: %s", cran_url)
                base.options(**kwargs)
                utils.install_packages(StrVector(names_to_install))
                # verify again to check if all dependencies are installed
                names_to_install = [
                    x for x in cran_dependencies if not rpackages.isinstalled(x)
                ]
                if len(names_to_install) > 0:
                    # conda package names
                    conda_packages = ["r-" + name.lower() for name in names_to_install]

                    conda_install_cmd = (
                        f"conda install -y -c conda-forge {' '.join(conda_packages)}"
                    )
                    logger.warning(
                        f"Failed to install the following CRAN dependencies: "
                        f"{names_to_install}. If you are using conda, you can try "
                        "installing the dependencies via:\n"
                        f"{conda_install_cmd}"
                    )

            else:
                cran_missing = names_to_install
        if bioconductor_dependencies:
            names_to_install = [
                x for x in bioconductor_dependencies if not rpackages.isinstalled(x)
            ]
            if len(names_to_install) > 0 and install_missing:
                logger.info(f"Using R_HOME: {r_home}")
                logger.info(
                    "Installing missing Bioconductor dependencies: "
                    f"{bioconductor_dependencies}"
                )
                if not rpackages.isinstalled("BiocManager"):
                    logger.info("Installing BiocManager")
                    utils.chooseCRANmirror(ind=1)
                    utils.install_packages("BiocManager")

                biocmanager = rpackages.importr("BiocManager")
                bioconductor_config_file, bioc_mirror = get_bioconductor_info()
                if cran_url is not None:
                    logger.info("Using CRAN mirror: %s", cran_url)
                if bioc_mirror is not None:
                    logger.info("Using Bioconductor mirror: %s", bioc_mirror)
                if bioc_mirror is not None and "BioC_mirror" not in kwargs:
                    kwargs["BioC_mirror"] = bioc_mirror
                if (
                    bioconductor_config_file is not None
                    and "BIOCONDUCTOR_CONFIG_FILE" not in kwargs
                ):
                    kwargs["BIOCONDUCTOR_CONFIG_FILE"] = bioconductor_config_file
                base.options(**kwargs)
                biocmanager.install(
                    StrVector(names_to_install), ask=False, update=False
                )
                # verify again to check if all dependencies are installed
                names_to_install = [
                    x for x in bioconductor_dependencies if not rpackages.isinstalled(x)
                ]
                if len(names_to_install) > 0:
                    conda_packages = [
                        "bioconductor-" + name.lower() for name in names_to_install
                    ]
                    conda_install_cmd = (
                        "conda install -y -c conda-forge -c bioconda "
                        f"{' '.join(conda_packages)}"
                    )
                    msg = (
                        f"Failed to install the following Bioconductor dependencies: "
                        f"{names_to_install}. If you are using conda, you can try "
                        "installing the dependencies via:\n"
                        f"conda install -y -c conda-forge {' '.join(conda_packages)}"
                    )
                    if sys.platform == "win32":
                        msg += (
                            "\nNote: Many of the Bioconductor packages are not "
                            "available on Windows. Please submit an issue on the BIOFIT "
                            "GitLab repository if you need help with installation."
                        )
                    logger.warning(msg)
            else:
                bioc_missing = names_to_install
        if (cran_missing or bioc_missing) and not install_missing:
            instructions = "Run `biofit.integration.R.RCaller.verify_r_dependencies("
            if cran_missing:
                instructions += f"cran_deps={cran_missing}, "
            if bioc_missing:
                instructions += f"bioc_deps={bioc_missing}, "
            instructions += "install_missing=True)` to install missing dependencies."
            raise PackageNotInstalledError(
                f"Missing R dependencies: {cran_missing + bioc_missing}. "
                f"{instructions}"
            )

    @classmethod
    def from_script(cls, r_code_or_path=None):
        self = cls()
        self._global_vars["R_SCRIPTS_PATH"] = (
            f"{config.R_SCRIPTS.resolve().as_posix()}/"
        )
        self.r_code = ""
        if r_code_or_path:
            if os.path.exists(r_code_or_path):
                self.r_source = r_code_or_path
                self._global_vars["CURRENT_DIR"] = (
                    f"{Path(self.r_source).resolve().parent.as_posix()}/"
                )
                with open(r_code_or_path, "r") as f:
                    self.r_code += f.read()
            else:
                self.r_code += r_code_or_path
        return self

    def _run_r(
        self,
        code_runner,
        r_code: str,
        env=None,
        add_globals=False,
        quiet=False,
    ) -> object:
        """
        Run R code using the provided code_runner function.

        Args:
            code_runner (Callable): The function to run the R code.
            r_code (str): The R code to run.
            env (SexpEnvironment, optional): The R environment to add global variables
                to. Defaults to globalenv.
            add_globals (bool, optional): Whether to add global variables to the R
                context. Defaults to False.
            quiet (bool, optional): Whether to suppress output from R. Defaults to
                False.

        Returns:
            The output from the R code or the converted output if convert is True.
        """
        from rpy2.rinterface_lib.embedded import RRuntimeError

        def run_code(env):
            if env is None:
                from rpy2.rinterface import globalenv

                env = globalenv
            if add_globals:
                for name, value in self._global_vars.items():
                    if value.endswith('"'):
                        value = value[:-1]
                    if value.startswith('"'):
                        value = value[1:]
                    env[name] = value
            results = code_runner(r_code)
            last_line = r_code.strip().split("\n")[-1].strip()
            # for some reason, arrow tables are not returned properly when letting
            # rpy2 handle the conversion. So we grab it directly from the R context.
            if last_line in env:
                results = env[last_line]
            return results

        if quiet:
            with ROutputCapture() as output_capture:
                try:
                    return run_code(env)
                except RRuntimeError as e:
                    # Attempt to capture the traceback from R.
                    r_stdout = output_capture.get_stdout()
                    r_stderr = output_capture.get_stderr()
                    logger.error(f"captured stdout from R: {r_stdout}")
                    logger.error(f"captured stderr from R: {r_stderr}")
                    try:
                        r_traceback = "\n".join(code_runner("unlist(traceback())"))
                    except Exception as traceback_exc:
                        r_traceback = f"Failed to capture R traceback: {traceback_exc}"
                    raise RuntimeError(f"Error in R code: {e}\n{r_traceback}") from e
        else:
            try:
                return run_code(env)
            except RRuntimeError as e:
                try:
                    r_traceback = "\n".join(code_runner("unlist(traceback())"))
                except Exception as traceback_exc:
                    r_traceback = f"Failed to capture R traceback: {traceback_exc}"
                raise RuntimeError(f"Error in R code: {e}\n{r_traceback}") from e

    def _convert_output(self, obj, code_runner, converter):
        # check if the result is a list
        is_list = converter.rpy2py(
            self._run_r(code_runner, "is.list", quiet=True)(obj)
        )[0]
        if not is_list:
            obj = [obj]

        names = self._run_r(code_runner, "names")(obj)
        has_no_names = converter.rpy2py(self._run_r(code_runner, "is.null")(names))[0]
        if has_no_names:
            names = [f"{i}" for i in range(len(obj))]
        out = {}

        for n, result in zip(names, obj):
            if converter.rpy2py(self._run_r(code_runner, "is.vector")(result))[0]:
                out[n] = np.array([r for r in result])
            elif converter.rpy2py(self._run_r(code_runner, "is.list")(result))[0]:
                out[n] = [r for r in result]
            elif result.__class__ in converter.rpy2py_nc_map and hasattr(
                result, "rclass"
            ):
                nc_map = converter.rpy2py_nc_map[result.__class__]
                for rclass in result.rclass:
                    if rclass in nc_map:
                        out[n] = nc_map[rclass](result)
                        break
                else:
                    out[n] = converter.rpy2py(result)
            else:
                out[n] = converter.rpy2py(result)

        if len(out) == 1:
            out = out[n]
        return out

    def _cleanup(self, code_runner):
        code_runner("rm(list=ls())")
        # garbage collection in Python doesn't always clean up after R automatically
        code_runner("gc()")
        gc.collect()

    def _init_vars(self, code_runner, env, converter, **kwargs):
        import rpy2.robjects as ro
        from rpy2.rinterface import Sexp

        def convert_to_r(arg):
            if isinstance(arg, Sexp):
                return arg
            elif arg is None:
                return ro.NULL
            elif isinstance(arg, (list, tuple)):
                return converter.py2rpy([convert_to_r(a) for a in arg])
            elif isinstance(arg, dict):
                return ro.ListVector(arg)
            else:
                return converter.py2rpy(arg)

        for n, arg in kwargs.items():
            env[n] = convert_to_r(arg)

    def get_method(self, name, enter_code="", exit_code="", convert=True, quiet=True):
        _converter = self._get_converter(name)
        import rpy2.rinterface
        import rpy2.robjects as ro

        with rpy2.rinterface.local_context() as r_context:
            self._run_r(ro.r, self.r_code, add_globals=True, quiet=quiet)
            func_args = list(_converter.rpy2py(r_context[name]).formals().names)
            self._cleanup(ro.r)

        def func(*args, context_kwargs: dict = None, **kwargs):
            if args:
                for n, arg in zip(func_args[: len(args)], args):
                    kwargs[n] = arg

            run_code = self.r_code + "\n" + enter_code + "\n"

            arg_str = ", ".join([f"{k} = {k}" for k in kwargs.keys() if k in func_args])

            # additional variables not in the function signature
            if context_kwargs:
                kwargs.update(context_kwargs)
            kwargs.update(self._global_vars)

            run_code += f"results <- {name}({arg_str})"

            run_code += "\n" + exit_code + "\nresults"

            out = self.run(run_code, convert=convert, quiet=quiet, **kwargs)

            return out

        parameters = [
            inspect.Parameter(name=arg, kind=inspect.Parameter.POSITIONAL_OR_KEYWORD)
            for arg in func_args
            if arg != "..."
        ]
        parameters += (
            [inspect.Parameter(name="kwargs", kind=inspect.Parameter.VAR_KEYWORD)]
            if "..." in func_args
            else []
        )
        func.__signature__ = inspect.Signature(parameters=parameters)
        return func

    def run(self, r_code=None, convert=True, quiet=False, **kwargs):
        """A method to run the R code with the provided arguments.

        Args:
            r_code (str, optional): The R code to run. Defaults to the provided R code
                given in the `from_script` method.
            convert (bool, optional): Whether to convert the R output to Python objects.
                Defaults to True.
            **kwargs: The variables to pass to the R code
        """
        r_code = r_code or self.r_code
        if r_code:
            r_code = r_code.strip()
            _converter = self._get_converter(self.run)

            import rpy2.rinterface
            import rpy2.robjects as ro

            with rpy2.rinterface.local_context() as r_context:
                self._init_vars(ro.r, r_context, _converter, **kwargs)
                out = self._run_r(ro.r, r_code, env=r_context, quiet=quiet)
                if out is None:
                    logger.warning("No output from R code")
                if convert:
                    if out is None:
                        raise RuntimeError(
                            "Conversion requested but no output from R code. Either "
                            "provide an output by adding the variable name on the last "
                            "line of the R code or set `convert` to `False`."
                        )
                    out = self._convert_output(out, ro.r, _converter)
                    self._cleanup(ro.r)
                return out
        else:
            raise RuntimeError(
                "R code not provided, please provide R code or the path to the R script"
                " via `from_script` method."
            )



from dataclasses import dataclass, field
from typing import List, Optional, Type

from biocore import DataHandler

from biofit.integration.biosets import get_feature
from biofit.processing import SelectedColumnTypes
from biocore.utils.py_util import is_bioset
from biofit.utils.types import Unset

from ..plot_scaling import ScalerPlotter, ScalerPlotterConfig


@dataclass
class CumulativeSumScalerPlotterConfig(ScalerPlotterConfig):
    processor_name: str = field(default="css", init=False, repr=False)
    _compare: bool = field(default=True, init=False, repr=False)
    _transoform_input_feature_types: List[Type] = field(
        default_factory=lambda: [
            None,
            None,
            get_feature("TARGET_FEATURE_TYPES"),
            get_feature("TARGET_FEATURE_TYPES"),
        ],
        init=False,
        repr=False,
    )

    _transform_unused_feature_types: List[Type] = field(
        default_factory=lambda: [
            get_feature("METADATA_FEATURE_TYPES"),
            get_feature("METADATA_FEATURE_TYPES"),
            None,
            None,
        ],
        init=False,
        repr=False,
    )

    input_columns1: Optional[str] = None
    input_columns2: Optional[str] = None
    label_name1: Optional[str] = None
    label_name2: Optional[str] = None
    ylab: Optional[str] = None
    xlab: Optional[str] = None
    ylim: Optional[list] = None
    before_title: str = "Before CSS"
    after_title: str = "After CSS"
    legend_position: str = "top"
    add_box: bool = True
    horizontal_plot: bool = False
    order: bool = False
    col_set: str = "Set1"
    cols: Optional[str] = None
    log_num: Optional[str] = None
    show_outliers: bool = True


@dataclass
class CumulativeSumScalerPlotterConfigForMetagenomics(CumulativeSumScalerPlotterConfig):
    _transform_input_feature_types: List[Type] = field(
        default_factory=lambda: [
            (get_feature("Abundance"), get_feature("ReadCount")),
            (get_feature("Abundance"), get_feature("ReadCount")),
            get_feature("TARGET_FEATURE_TYPES"),
            get_feature("TARGET_FEATURE_TYPES"),
        ],
        init=False,
        repr=False,
    )
    dataset_name: str = field(default="metagenomics", init=False, repr=False)
    log_num: Optional[str] = "log2_1p"


@dataclass
class CumulativeSumScalerPlotterConfigForOTU(CumulativeSumScalerPlotterConfig):
    _transform_input_feature_types: List[Type] = field(
        default_factory=lambda: [
            get_feature("Abundance"),
            get_feature("Abundance"),
            get_feature("TARGET_FEATURE_TYPES"),
            get_feature("TARGET_FEATURE_TYPES"),
        ],
        init=False,
        repr=False,
    )
    dataset_name: str = field(default="otu", init=False, repr=False)
    ylab: Optional[str] = "OTU Abundance"
    log_num: Optional[str] = "log10_1p"


@dataclass
class CumulativeSumScalerPlotterConfigForASV(CumulativeSumScalerPlotterConfig):
    _transform_input_feature_types: List[Type] = field(
        default_factory=lambda: [
            get_feature("Abundance"),
            get_feature("Abundance"),
            get_feature("TARGET_FEATURE_TYPES"),
            get_feature("TARGET_FEATURE_TYPES"),
        ],
        init=False,
        repr=False,
    )
    dataset_name: str = field(default="asv", init=False, repr=False)


@dataclass
class CumulativeSumScalerPlotterConfigForGenomics(CumulativeSumScalerPlotterConfig):
    _fit_input_feature_types: List[Type] = field(
        default_factory=lambda: [
            get_feature("GenomicVariant"),
            get_feature("GenomicVariant"),
            get_feature("TARGET_FEATURE_TYPES"),
            get_feature("TARGET_FEATURE_TYPES"),
        ],
        init=False,
        repr=False,
    )
    _transform_input_feature_types: List[Type] = field(
        default_factory=lambda: [
            get_feature("GenomicVariant"),
            get_feature("GenomicVariant"),
            get_feature("TARGET_FEATURE_TYPES"),
            get_feature("TARGET_FEATURE_TYPES"),
        ],
        init=False,
        repr=False,
    )
    dataset_name: str = field(default="genomics", init=False, repr=False)


@dataclass
class CumulativeSumScalerPlotterConfigForSNP(CumulativeSumScalerPlotterConfig):
    _fit_input_feature_types: List[Type] = field(
        default_factory=lambda: [
            get_feature("GenomicVariant"),
            get_feature("GenomicVariant"),
            get_feature("TARGET_FEATURE_TYPES"),
            get_feature("TARGET_FEATURE_TYPES"),
        ],
        init=False,
        repr=False,
    )
    _transform_input_feature_types: List[Type] = field(
        default_factory=lambda: [
            get_feature("GenomicVariant"),
            get_feature("GenomicVariant"),
            get_feature("TARGET_FEATURE_TYPES"),
            get_feature("TARGET_FEATURE_TYPES"),
        ],
        init=False,
        repr=False,
    )
    dataset_name: str = field(default="snp", init=False, repr=False)


@dataclass
class CumulativeSumScalerPlotterConfigForReadCount(CumulativeSumScalerPlotterConfig):
    _fit_input_feature_types: List[Type] = field(
        default_factory=lambda: [
            get_feature("ReadCount"),
            get_feature("ReadCount"),
            get_feature("TARGET_FEATURE_TYPES"),
            get_feature("TARGET_FEATURE_TYPES"),
        ],
        init=False,
        repr=False,
    )
    _transform_input_feature_types: List[Type] = field(
        default_factory=lambda: [
            get_feature("ReadCount"),
            get_feature("ReadCount"),
            get_feature("TARGET_FEATURE_TYPES"),
            get_feature("TARGET_FEATURE_TYPES"),
        ],
        init=False,
        repr=False,
    )
    dataset_name: str = field(default="read_count", init=False, repr=False)


class CumulativeSumScalerPlotter(ScalerPlotter):
    config_class = CumulativeSumScalerPlotterConfig
    config: CumulativeSumScalerPlotterConfig

    def __init__(
        self,
        ylab: Optional[str] = Unset("None"),
        xlab: Optional[str] = Unset("None"),
        ylim: Optional[list] = Unset("None"),
        title: str = Unset('"Violin Plot"'),
        legend_position: str = Unset('"top"'),
        add_box: bool = Unset("True"),
        horizontal_plot: bool = Unset("False"),
        order: bool = Unset("False"),
        col_set: str = Unset('"Set1"'),
        cols: Optional[str] = Unset("None"),
        log_num: Optional[int] = Unset("None"),
        show_outliers: bool = Unset("True"),
        config: CumulativeSumScalerPlotterConfig = None,
        **kwargs,
    ):
        super().__init__(
            config=config,
            ylab=ylab,
            xlab=xlab,
            ylim=ylim,
            title=title,
            legend_position=legend_position,
            add_box=add_box,
            horizontal_plot=horizontal_plot,
            order=order,
            col_set=col_set,
            cols=cols,
            log_num=log_num,
            show_outliers=show_outliers,
            **kwargs,
        )

    def plot(
        self,
        x1,
        x2=None,
        y1=None,
        y2=None,
        input_columns1: SelectedColumnTypes = None,
        input_columns2: SelectedColumnTypes = None,
        label_name1: SelectedColumnTypes = None,
        label_name2: SelectedColumnTypes = None,
        ylab: Optional[str] = Unset("None"),
        xlab: Optional[str] = Unset("None"),
        ylim: Optional[list] = Unset("None"),
        title: str = Unset('"Violin Plot"'),
        legend_position: str = Unset('"top"'),
        add_box: bool = Unset("True"),
        horizontal_plot: bool = Unset("False"),
        order: bool = Unset("False"),
        col_set: str = Unset('"Set1"'),
        cols: Optional[str] = Unset("None"),
        log_num: Optional[int] = Unset("None"),
        show_outliers: bool = Unset("True"),
        path: str = None,
        device: str = "pdf",
        fingerprint: str = None,
        unused_columns: SelectedColumnTypes = None,
        raise_if_missing: bool = True,
        show: bool = True,
    ):
        if is_bioset(x2):
            from biosets import get_target

            if y2 is None:
                y2 = get_target(x2)
        if x1 is None or x2 is None:
            raise ValueError("Must provide the before and after normalization.")
        if y1 is not None and y2 is None:
            raise ValueError("Must provide the target for the after normalization.")
        if y1 is None:
            if label_name1 is None:
                raise ValueError(
                    "Must provide the target for the before normalization."
                )
            y1 = DataHandler.select_columns(x1, label_name1)
            x1 = DataHandler.drop_columns(x1, label_name1)
        if y2 is None:
            if label_name2 is None:
                raise ValueError("Must provide the target for the after normalization.")
            y2 = DataHandler.select_columns(x2, label_name2)
            x2 = DataHandler.drop_columns(x2, label_name2)

        self.config._input_columns = self._set_input_columns_and_arity(
            input_columns1, input_columns2, label_name1, label_name2
        )
        return self._plot(
            x1,
            x2,
            y1,
            y2,
            ylab=ylab,
            xlab=xlab,
            ylim=ylim,
            title=title,
            legend_position=legend_position,
            add_box=add_box,
            horizontal_plot=horizontal_plot,
            order=order,
            col_set=col_set,
            cols=cols,
            log_num=log_num,
            show_outliers=show_outliers,
            path=path,
            device=device,
            fingerprint=fingerprint,
            unused_columns=unused_columns,
            raise_if_missing=raise_if_missing,
            show=show,
        )



from dataclasses import dataclass, field

from biofit.processing import BaseProcessor, ProcessorConfig, SelectedColumnTypes


@dataclass
class StatConfig(ProcessorConfig):
    processor_type: str = field(default="stat", init=False, repr=False)


class Stat(BaseProcessor):
    """Base class for statistical processors."""

    def transform(
        self,
        X,
        input_columns: SelectedColumnTypes = None,
        keep_unused_columns: bool = None,
        raise_if_missing: bool = None,
        cache_output: bool = None,
        cache_dir: str = None,
        cache_file_name: str = None,
        load_from_cache_file: bool = None,
        batched: bool = None,
        batch_size: int = None,
        batch_format: str = None,
        output_format: str = None,
        map_kwargs: dict = None,
        num_proc: int = None,
        fingerprint: str = None,
    ):
        self._input_columns = self._set_input_columns_and_arity(input_columns)
        return self._process_transform(
            X,
            keep_unused_columns=keep_unused_columns,
            raise_if_missing=raise_if_missing,
            cache_output=cache_output,
            cache_dir=cache_dir,
            cache_file_name=cache_file_name,
            load_from_cache_file=load_from_cache_file,
            batched=batched,
            batch_size=batch_size,
            batch_format=batch_format,
            output_format=output_format,
            map_kwargs=map_kwargs,
            num_proc=num_proc,
            fingerprint=fingerprint,
        )

    def fit_transform(
        self,
        X,
        *args,
        input_columns: SelectedColumnTypes = None,
        keep_unused_columns: bool = True,
        raise_if_missing: bool = False,
        cache_output: bool = None,
        load_from_cache_file: bool = None,
        batched: bool = True,
        batch_size: int = 1000,
        batch_format: str = None,
        output_format: str = None,
        map_kwargs: dict = {"fn_kwargs": {}},
        num_proc: int = None,
        cache_dir: str = None,
        cache_file_name: str = None,
        fingerprint: str = None,
    ):
        return self.fit(
            X,
            input_columns=input_columns,
            raise_if_missing=raise_if_missing,
            cache_output=cache_output,
            load_from_cache_file=load_from_cache_file,
            batched=batched,
            batch_size=batch_size,
            batch_format=batch_format,
            num_proc=num_proc,
            map_kwargs=map_kwargs,
            cache_dir=cache_dir,
            cache_file_name=cache_file_name,
            fingerprint=fingerprint,
        ).transform(
            X,
            input_columns=input_columns,
            keep_unused_columns=keep_unused_columns,
            raise_if_missing=raise_if_missing,
            cache_output=cache_output,
            cache_dir=cache_dir,
            cache_file_name=cache_file_name,
            load_from_cache_file=load_from_cache_file,
            batched=batched,
            batch_size=batch_size,
            batch_format=batch_format,
            output_format=output_format,
            map_kwargs=map_kwargs,
            num_proc=num_proc,
            fingerprint=fingerprint,
        )

    def _process_transform(self, *args, **kwargs):
        kwargs["keep_unused_columns"] = False
        return super()._process_transform(*args, **kwargs)
